/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if !defined(WIN32) && !defined(_WIN32_WCE)
#ifdef __APPLE__
#include <sys/types.h>
#endif
#include <sys/socket.h>
#include <netdb.h>
#endif

#include "mediastreamer2/msticker.h"
#include "mediastreamer2/ice.h"
#include "mediastreamer2/mscommon.h"

struct IceData {
  RtpSession *session;
  struct IceCheckList *check_lists;	/* table of 10 cpair */
  int rate;
  uint64_t last_packet_sent;
  int last_packet_sent_tick;
  int stun_indication_received;
  int delay_for_first_stun_indication;
};

typedef struct IceData IceData_t;

static void ice_init(MSFilter * f)
{
  IceData_t *d = (IceData_t *)ms_new(IceData_t, 1);

  d->session = NULL;
  d->check_lists = NULL;
  d->last_packet_sent = 0;
  d->last_packet_sent_tick = -1;
  d->stun_indication_received = 0;
  d->delay_for_first_stun_indication = 0;

  f->data = d;
}

static void ice_postprocess(MSFilter * f)
{
  IceData_t *d = (IceData_t *) f->data;
  d->last_packet_sent = 0;
  d->last_packet_sent_tick = -1;
  d->stun_indication_received = 0;
}

static void ice_uninit(MSFilter * f)
{
  IceData_t *d = (IceData_t *) f->data;

  if (d->check_lists!=NULL && d->check_lists->cand_pairs!=NULL) {
    OList *elem;
    for (elem=d->check_lists->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      ortp_free(cd);
    }
    d->check_lists->cand_pairs = o_list_free(d->check_lists->cand_pairs);
  }
  if (d->check_lists!=NULL)
    ortp_free(d->check_lists);

  ms_free(f->data);
}

static int ice_set_session(MSFilter * f, void *arg)
{
  IceData_t *d = (IceData_t *) f->data;
  RtpSession *s = (RtpSession *) arg;
  d->session = s;

  return 0;
}

static int ice_set_sdpcandidates(MSFilter * f, void *arg)
{
  IceData_t *d = (IceData_t *) f->data;
  struct IceCheckList *scs = NULL;

  if (d == NULL)
    return -1;

  if (d->check_lists!=NULL && d->check_lists->cand_pairs!=NULL) {
    OList *elem;
    for (elem=d->check_lists->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      ortp_free(cd);
    }
    d->check_lists->cand_pairs = o_list_free(d->check_lists->cand_pairs);
  }
  if (d->check_lists!=NULL)
    ortp_free(d->check_lists);

  d->check_lists=ortp_new0(struct IceCheckList,1);

  scs = (struct IceCheckList *) arg;
  if (scs!=NULL) {
    OList *elem;

    memcpy(d->check_lists, scs, sizeof(struct IceCheckList));
    d->check_lists->cand_pairs=NULL; /* build a duplicated list */

    for (elem=scs->cand_pairs;elem!=NULL;elem=o_list_next(elem)){
      struct CandidatePair *cd=(struct CandidatePair*) elem->data;
      struct CandidatePair *cd_dup=ortp_new(struct CandidatePair,1);
      memcpy(cd_dup, cd, sizeof(struct CandidatePair));
      d->check_lists->cand_pairs = o_list_append(d->check_lists->cand_pairs, cd_dup);
    }
  }

  return 0;
}

static int ice_set_delay_for_stun_indication(MSFilter * f, void *arg) {
  IceData_t *d = (IceData_t *)f->data;

  if (d == NULL)
    return -1;

  d->delay_for_first_stun_indication = *((int*)arg);
  return 0;
}

static void ice_preprocess(MSFilter * f){
  IceData_t *d = (IceData_t *) f->data;
  d->last_packet_sent = 0;
  d->last_packet_sent_tick = -1;
}

static void ice_process(MSFilter * f)
{
  IceData_t *d = (IceData_t *) f->data;

  if (d->session == NULL)
    return;

  if (d->delay_for_first_stun_indication == 10)
	  return;

  if (d->last_packet_sent_tick < 0)
  {
    d->last_packet_sent = d->session->rtp.stats.packet_sent;
    /* start with a stun-indication anyway, except if delay_for_first_stun_indication>0 */
    d->last_packet_sent_tick = f->ticker->ticks-100+d->delay_for_first_stun_indication;
  }

  if (f->ticker->ticks-d->last_packet_sent_tick==100) {

    /* if no rtp sent, send a STUN INDICATION OVER THE SOCKET */
    /* This is a keep alive mechanism described in rfc6263 with no */
    /* impact on remote agent */
    if (d->last_packet_sent==d->session->rtp.stats.packet_sent) {
      StunMessage req;
      int error;
      char buf[STUN_MAX_MESSAGE_SIZE];
      int len = STUN_MAX_MESSAGE_SIZE;
      struct sockaddr *destaddr=(struct sockaddr*)&d->session->rtp.rem_addr;
      ortp_socket_t sockfd=0;
      memset(&req, 0, sizeof(StunMessage));
      stunBuildReqSimple( &req, NULL, FALSE, FALSE, 1);
      req.msgHdr.msgType = (STUN_METHOD_BINDING|STUN_INDICATION);
      req.hasFingerprint = TRUE;
      len = stunEncodeMessage( &req, buf, len, NULL);

#if defined(WIN32) || defined(_WIN32_WCE)
#define sockisvalid(s) (s != ~0)
#else
#define sockisvalid(s) ((s) >= 0)
#endif
      {
        int idx;
        for (idx=0;sockisvalid(d->session->rtp.socketv4[idx]);idx++) {
          if (d->session->rtp.sockfamily[idx]==destaddr->sa_family) {
            sockfd=d->session->rtp.socketv4[idx];
            break;
          }
        }
      }

      error = sendto (sockfd, buf, len,
        0,destaddr, d->session->rtp.rem_addrlen);
      if (error < 0)
        ortp_warning ("Error sending stun indication packet: socket=%i", sockfd);
    }

    d->last_packet_sent_tick=f->ticker->ticks;
    d->last_packet_sent=d->session->rtp.stats.packet_sent;
  }
}

static int get_statistics(MSFilter *f, void *arg){
  int i;
  ms_warning("filter: %s[%i->%i]", f->desc->name,
    f->desc->ninputs, f->desc->noutputs);
  for (i=0;i<f->desc->ninputs;i++) {
    if (f->inputs[i]!=NULL) ms_warning("filter: %s in[%i]=%i", f->desc->name,
      i, f->inputs[i]->q.q_mcount);
  }
  for (i=0;i<f->desc->noutputs;i++) {
    if (f->outputs[i]!=NULL) ms_warning("filter: %s out[%i]=%i", f->desc->name,
      i, f->outputs[i]->q.q_mcount);
  }
  return 0;
}

static MSFilterMethod ice_methods[] = {
  {MS_ICE_SET_SESSION, ice_set_session},
  {MS_ICE_SET_CANDIDATEPAIRS, ice_set_sdpcandidates},
  {MS_ICE_SET_DELAY_FOR_STUN_INDICATION, ice_set_delay_for_stun_indication},
  {MS_FILTER_GET_STATISTICS, get_statistics },
  {0, NULL}
};

#ifdef _MSC_VER

MSFilterDesc ms_ice_desc = {
  MS_ICE_ID,
  "MSIce",
  N_("ICE filter"),
  MS_FILTER_OTHER,
  NULL,
  0,
  0,
  ice_init,
  ice_preprocess,
  ice_process,
  ice_postprocess,
  ice_uninit,
  ice_methods
};

#else

MSFilterDesc ms_ice_desc = {
  .id = MS_ICE_ID,
  .name = "MSIce",
  .text = N_("ICE filter"),
  .category = MS_FILTER_OTHER,
  .ninputs = 0,
  .noutputs = 0,
  .init = ice_init,
  .preprocess = ice_preprocess,
  .process = ice_process,
  .postprocess=ice_postprocess,
  .uninit = ice_uninit,
  .methods = ice_methods
};

#endif

MS_FILTER_DESC_EXPORT(ms_ice_desc)
