/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef __APPLE__
#include "TargetConditionals.h"
#endif

#include "mediastreamer2/msfilter.h"

#ifdef _MSC_VER
#include <malloc.h>
#endif

#include <speex/speex_resampler.h>
#include <math.h>

typedef struct _ResampleData{
	MSBufferizer *bz;
	uint32_t ts;
	uint32_t input_rate;
	uint32_t output_rate;

	SpeexResamplerState *handle;
	int nb_unprocessed;
} ResampleData;

static ResampleData * resample_data_new(){
	ResampleData *obj=(ResampleData *)ms_new(ResampleData,1);
	obj->bz=ms_bufferizer_new();
	obj->ts=0;
	obj->input_rate=8000;
	obj->output_rate=16000;
	obj->handle=NULL;

	obj->nb_unprocessed=0;
	return obj;
}

static void resample_data_destroy(ResampleData *obj){
	if (obj->handle!=NULL)
		speex_resampler_destroy(obj->handle);
	ms_bufferizer_destroy(obj->bz);
	ms_free(obj);
}

static void resample_init(MSFilter *obj){
	obj->data=resample_data_new();
}

static void resample_uninit(MSFilter *obj){
	resample_data_destroy((ResampleData*)obj->data);
}

static void resample_process(MSFilter *obj){
	ResampleData *dt=(ResampleData*)obj->data;
	MSBufferizer *bz=dt->bz;
	uint8_t buffer[7680]; /* [8k -> 192k] */
	int size_of_input;
	int size_of_output;

	mblk_t *m;
	int new_input_rate;

	while ((m = ms_queue_get(obj->inputs[0])) != NULL) {

		new_input_rate = mblk_get_rate(m);
		if (new_input_rate > 0 && new_input_rate!= dt->input_rate) {
			if (dt->handle != NULL) {
				speex_resampler_destroy(dt->handle);
				dt->handle = 0;
			}
			dt->input_rate = new_input_rate;
			ms_bufferizer_flush(bz);
		}
		if (dt->output_rate == dt->input_rate) {
			ms_queue_put(obj->outputs[0], m);
			continue;
		}

		/* rates are different: convert */
		if (dt->handle != NULL) {
			unsigned int inrate = 0, outrate = 0;
			speex_resampler_get_rate(dt->handle, &inrate, &outrate);
			if (inrate != dt->input_rate || outrate != dt->output_rate) {
				speex_resampler_destroy(dt->handle);
				dt->handle = 0;
				ms_bufferizer_flush(bz);
			}
		}
		if (dt->handle == NULL) {
			int err = 0;
			int quality = SPEEX_RESAMPLER_QUALITY_VOIP;
#if (TARGET_OS_IPHONE == 1) || (TARGET_IPHONE_SIMULATOR == 1)
			quality = 0;
#elif defined(ANDROID)
			quality = 1;
#elif defined(__arm__) || defined(_M_ARM)
			quality = 0;
#endif
			dt->handle = speex_resampler_init(1, dt->input_rate, dt->output_rate, SPEEX_RESAMPLER_QUALITY_VOIP, &err);
		}

		if (dt->input_rate < dt->output_rate)
			size_of_input = 320 * dt->input_rate / 8000;
		else
			size_of_input = 320 * dt->input_rate / 8000;
		size_of_output = (size_of_input * dt->output_rate) / dt->input_rate;

		ms_bufferizer_put(bz, m);

		while (ms_bufferizer_read(bz, buffer, size_of_input) == size_of_input) {
			mblk_t *obl = allocb(size_of_output, 0);

			spx_uint32_t in_len;
			spx_uint32_t out_len;

			in_len = size_of_input / 2;
			out_len = size_of_output / 2;
			speex_resampler_process_int(dt->handle, 0, (short*)buffer, &in_len, (spx_int16_t*)obl->b_wptr, &out_len);

			if (in_len != size_of_input / 2) {
				ms_debug("resampling info: in_len=%i, out_len=%i", in_len, out_len);
			}

			obl->b_wptr = obl->b_wptr + (out_len * 2); /* size_of_output; */

			mblk_set_timestamp_info(obl, dt->ts);
			dt->ts += 160;

			ms_queue_put(obj->outputs[0], obl);
		}
	}
}

static int ms_resample_set_sr(MSFilter *obj, void *arg){
  ResampleData *dt=(ResampleData*)obj->data;
  dt->input_rate=((int*)arg)[0];
  return 0;
}

static int ms_resample_set_output_sr(MSFilter *obj, void *arg){
  ResampleData *dt=(ResampleData*)obj->data;
  dt->output_rate=((int*)arg)[0];
  return 0;
}

static int get_statistics(MSFilter *f, void *arg){
	ResampleData *dt=(ResampleData*)f->data;
	ms_warning("filter: %s[%i->%i]", f->desc->name,
		f->desc->ninputs, f->desc->noutputs);
	if (f->inputs[0]!=NULL) ms_warning("filter: %s in[0]=%i", f->desc->name,
		f->inputs[0]->q.q_mcount);
	if (f->outputs[0]!=NULL) ms_warning("filter: %s out[0]=%i", f->desc->name,
		f->outputs[0]->q.q_mcount);
	ms_warning("filter: %s bz[0]=%i", f->desc->name,
			ms_bufferizer_get_avail(dt->bz));
	return 0;
}

static MSFilterMethod enc_methods[]={
  {	MS_FILTER_SET_SAMPLE_RATE	 ,	ms_resample_set_sr		},
  {	MS_FILTER_SET_OUTPUT_SAMPLE_RATE ,	ms_resample_set_output_sr	},
  {	MS_FILTER_GET_STATISTICS, get_statistics },
  {	0				 ,	NULL	}
};

#ifdef _MSC_VER

MSFilterDesc ms_resample_desc={
	MS_RESAMPLE_ID,
	"MSResample",
	N_("frequency resampler"),
	MS_FILTER_OTHER,
	NULL,
	1,
	1,
	resample_init,
	NULL,
	resample_process,
	NULL,
	resample_uninit,
	enc_methods
};

#else

MSFilterDesc ms_resample_desc={
	.id=MS_RESAMPLE_ID,
	.name="MSResample",
	.text=N_("frequency resampler"),
	.category=MS_FILTER_OTHER,
	.ninputs=1,
	.noutputs=1,
	.init=resample_init,
	.process=resample_process,
	.uninit=resample_uninit,
	.methods=enc_methods
};

#endif

MS_FILTER_DESC_EXPORT(ms_resample_desc)

