/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "mediastreamer2/msrtp.h"
#include "mediastreamer2/msticker.h"

#include "ortp/telephonyevents.h"
#if defined(__cplusplus)
#define B64_NO_NAMESPACE
#endif
#include "ortp/b64.h"


struct _SenderData {
  RtpSession *session;
  uint32_t tsoff;
  uint32_t skip_until;
  int rate;
  char dtmf;
  bool_t dtmf_start;
  int dtmf_duration;
  char relay_session_id[64];
  int relay_session_id_size;
  uint64_t last_rsi_time;
  bool_t skip;
  bool_t mute_mic;
};

typedef struct _SenderData SenderData;

static void sender_init(MSFilter * f)
{
  SenderData *d = (SenderData *)ms_new(SenderData, 1);

  d->session = NULL;
  d->tsoff = 0;
  d->skip_until = 0;
  d->skip = FALSE;
  d->rate = 8000;
  d->dtmf = 0;
  d->dtmf_start = FALSE;
  d->dtmf_duration = 800;
  d->mute_mic=FALSE;
  d->relay_session_id_size=0;
  d->last_rsi_time=0;
  f->data = d;
}

static void sender_uninit(MSFilter * f)
{
  SenderData *d = (SenderData *) f->data;

  ms_free(d);
}

static int sender_send_dtmf(MSFilter * f, void *arg)
{
  const char *dtmf = (const char *) arg;
  SenderData *d = (SenderData *) f->data;

  ms_filter_lock(f);
  if (d->skip==TRUE)
  {
    ms_filter_unlock(f);
    return -1;
  }
  d->dtmf = dtmf[0];
  ms_filter_unlock(f);
  return 0;
}

static int sender_set_dtmf_duration(MSFilter * f, void *arg)
{
  SenderData *d = (SenderData *) f->data;
  d->dtmf_duration = *((int*)arg);
  return 0;
}

static int sender_set_session(MSFilter * f, void *arg)
{
  SenderData *d = (SenderData *) f->data;
  RtpSession *s = (RtpSession *) arg;
  PayloadType *pt =
    rtp_profile_get_payload(rtp_session_get_profile(s),
    rtp_session_get_send_payload_type(s));
  if (pt != NULL) {
    if (strcasecmp("g722", pt->mime_type)==0 )
      d->rate=8000;
    else d->rate = pt->clock_rate;
  } else {
    pt = rtp_profile_get_payload(rtp_session_get_profile(s), rtp_session_get_recv_payload_type(s));
    if (pt!=NULL) {
      ms_message("checking for recv payload clockrate");
      if (strcasecmp("g722", pt->mime_type)==0 )
        d->rate=8000;
      else d->rate = pt->clock_rate;
    } else {
      ms_warning("Sending undefined payload type ?");
    }
  }
  d->session = s;
  return 0;
}

static int sender_mute_mic(MSFilter * f, void *arg)
{
  SenderData *d = (SenderData *) f->data;
  ms_filter_lock(f);
  d->mute_mic=TRUE;
  ms_filter_unlock(f);
  return 0;
}

static int sender_unmute_mic(MSFilter * f, void *arg)
{
  SenderData *d = (SenderData *) f->data;
  ms_filter_lock(f);
  d->mute_mic=FALSE;
  ms_filter_unlock(f);
  return 0;
}

static int sender_set_relay_session_id(MSFilter *f, void*arg){
  SenderData *d = (SenderData *) f->data;
  const char *tmp=(const char *)arg;
  d->relay_session_id_size=(int)b64_decode(tmp, strlen(tmp), (void*)d->relay_session_id, (unsigned int)sizeof(d->relay_session_id));
  return 0;
}

static int sender_get_sr(MSFilter *f, void *arg){
  SenderData *d = (SenderData *) f->data;
  *(int*)arg=d->rate;
  return 0;
}

/* the goal of that function is to return a absolute timestamp closest to real time, with respect of given packet_ts, which is a relative to an undefined origin*/
static uint32_t get_cur_timestamp(MSFilter * f, uint32_t packet_ts)
{
  SenderData *d = (SenderData *) f->data;
  uint32_t curts = (uint32_t)( f->ticker->time*(uint64_t)d->rate/(uint64_t)1000) ;
  int diff;
  int delta = d->rate / 50;	/*20 ms at 8000Hz */
  uint32_t netts;

  netts = packet_ts + d->tsoff;
  diff = curts - netts;

  if (diff > delta) {
    d->tsoff = curts - packet_ts;
    netts = packet_ts + d->tsoff;
    ms_message("synchronizing timestamp, diff=%i", diff);
  }
  else if (diff < -(delta * 5)) {
    d->tsoff = curts - packet_ts;
    netts = packet_ts + d->tsoff;
    ms_message("ticker back in time diff=%i", diff);
    /* hardware clock is going slower than sound card on my PDA... */
  }

  return netts;
}

static int send_dtmf(MSFilter * f, uint32_t timestamp_start, uint32_t current_timestamp)
{
  SenderData *d = (SenderData *) f->data;
  mblk_t *m1;
  int tev_type;

  /* create the first telephony event packet */
  switch (d->dtmf){
  case '1':
    tev_type=TEV_DTMF_1;
    break;
  case '2':
    tev_type=TEV_DTMF_2;
    break;
  case '3':
    tev_type=TEV_DTMF_3;
    break;
  case '4':
    tev_type=TEV_DTMF_4;
    break;
  case '5':
    tev_type=TEV_DTMF_5;
    break;
  case '6':
    tev_type=TEV_DTMF_6;
    break;
  case '7':
    tev_type=TEV_DTMF_7;
    break;
  case '8':
    tev_type=TEV_DTMF_8;
    break;
  case '9':
    tev_type=TEV_DTMF_9;
    break;
  case '*':
    tev_type=TEV_DTMF_STAR;
    break;
  case '0':
    tev_type=TEV_DTMF_0;
    break;
  case '#':
    tev_type=TEV_DTMF_POUND;
    break;

  case 'A':
  case 'a':
    tev_type=TEV_DTMF_A;
    break;


  case 'B':
  case 'b':
    tev_type=TEV_DTMF_B;
    break;

  case 'C':
  case 'c':
    tev_type=TEV_DTMF_C;
    break;

  case 'D':
  case 'd':
    tev_type=TEV_DTMF_D;
    break;

  case '!':
    tev_type=TEV_FLASH;
    break;


  default:
    ms_warning("Bad dtmf: %c.",d->dtmf);
    return -1;
  }


  if (d->dtmf_start == TRUE)
    m1=rtp_session_create_telephone_event_packet(d->session,1);
  else
    m1=rtp_session_create_telephone_event_packet(d->session,0);
  if (m1==NULL) return -1;


  if (RTP_TIMESTAMP_IS_NEWER_THAN(current_timestamp, d->skip_until)) {
    //retransmit end of rtp dtmf event
    mblk_t *tmp;
    rtp_session_add_telephone_event(d->session,m1,tev_type,1,10, (current_timestamp-timestamp_start));
    tmp=copymsg(m1);
    rtp_session_sendm_with_ts(d->session,tmp,timestamp_start+160);
    d->session->rtp.snd_seq--;
    tmp=copymsg(m1);
    rtp_session_sendm_with_ts(d->session,tmp,timestamp_start+160);
    d->session->rtp.snd_seq--;
    rtp_session_sendm_with_ts(d->session,m1,timestamp_start+160);
  }
  else {
    rtp_session_add_telephone_event(d->session,m1,tev_type,0,10, (current_timestamp-timestamp_start));
    rtp_session_sendm_with_ts(d->session,m1,timestamp_start+160);
  }
  return 0;
}

static void sender_postprocess(MSFilter * f)
{
  SenderData *d = (SenderData *) f->data;
  if (d->session!=NULL)
    rtp_session_disable_sendrecv(d->session);
}

static void sender_process(MSFilter * f)
{
  SenderData *d = (SenderData *) f->data;
  RtpSession *s = d->session;

  mblk_t *im;
  uint32_t timestamp;
  int burst;

  if (s == NULL){
    ms_queue_flush(f->inputs[0]);
    return;
  }
  
  if (d->session->bundle!=NULL) rtp_bundle_lock(d->session->bundle);
  rtp_session_enable_sendrecv (d->session);

#if 0
  if (f->inputs[0]->q.q_mcount>10)
  {
    ms_warning("RTP burst to be sent: %i", f->inputs[0]->q.q_mcount);
  }
#endif

  if (d->relay_session_id_size>0 && 
    ( (f->ticker->time-d->last_rsi_time)>5000 || d->last_rsi_time==0) ) {
      ms_message("relay session id sent in RTCP APP");
      rtp_session_send_rtcp_APP(s,0,"RSID",(const uint8_t *)d->relay_session_id,d->relay_session_id_size);
      d->last_rsi_time=f->ticker->time;
  }
  burst = 4;
  while ((im = ms_queue_get(f->inputs[0])) != NULL) {
    mblk_t *header;

    timestamp = get_cur_timestamp(f, mblk_get_timestamp_info(im));
    ms_filter_lock(f);
    if (d->skip) {
      ms_debug("skipping..");
      send_dtmf(f, d->skip_until-d->dtmf_duration, timestamp);
      d->dtmf_start = FALSE;
      if (!RTP_TIMESTAMP_IS_NEWER_THAN(timestamp, d->skip_until)) {
        freemsg(im);
        ms_filter_unlock(f);
        continue;
      }
      d->skip = FALSE;
      d->dtmf = 0;
    }

    if (d->skip == FALSE && d->mute_mic==FALSE){
      int pt = mblk_get_payload_type(im);
      header = rtp_session_create_packet(s, 12, NULL, 0);
      if (pt>0)
        rtp_set_payload_type(header, pt);
      rtp_set_markbit(header, mblk_get_marker_info(im));
      header->b_cont = im;
      rtp_session_sendm_with_ts(s, header, timestamp);
    }
    else{
      freemsg(im);
    }

    if (d->dtmf != 0) {
      ms_debug("prepare to send RFC2833 dtmf.");
      d->skip_until = timestamp + d->dtmf_duration;
      d->skip = TRUE;
      d->dtmf_start = TRUE;
    }
    ms_filter_unlock(f);

    burst--;
    if (burst <= 0)
    {
      ms_usleep(2000);
      burst=4;
    }
  }
  if (d->session->bundle!=NULL) rtp_bundle_unlock(d->session->bundle);
}

static int sender_get_statistics(MSFilter *f, void *arg){
  int i;
  ms_warning("filter: %s[%i->%i]", f->desc->name,
    f->desc->ninputs, f->desc->noutputs);
  for (i=0;i<f->desc->ninputs;i++) {
    if (f->inputs[i]!=NULL) ms_warning("filter: %s in[%i]=%i", f->desc->name,
      i, f->inputs[i]->q.q_mcount);
  }
  for (i=0;i<f->desc->noutputs;i++) {
    if (f->outputs[i]!=NULL) ms_warning("filter: %s out[%i]=%i", f->desc->name,
      i, f->outputs[i]->q.q_mcount);
  }
  return 0;
}

static MSFilterMethod sender_methods[] = {
  {MS_RTP_SEND_MUTE_MIC, sender_mute_mic},
  {MS_RTP_SEND_UNMUTE_MIC, sender_unmute_mic},
  {MS_RTP_SEND_SET_SESSION, sender_set_session},
  {MS_RTP_SEND_SEND_DTMF, sender_send_dtmf},
  {MS_RTP_SEND_SET_RELAY_SESSION_ID, sender_set_relay_session_id},
  {MS_FILTER_GET_SAMPLE_RATE, sender_get_sr },
  {MS_RTP_SEND_SET_DTMF_DURATION, sender_set_dtmf_duration },
  {MS_FILTER_GET_STATISTICS, sender_get_statistics },
  {0, NULL}
};

MSFilterDesc ms_rtp_send_desc = {
  MS_RTP_SEND_ID,
  "MSRtpSend",
  N_("RTP output filter"),
  MS_FILTER_OTHER,
  NULL,
  1,
  0,
  sender_init,
  NULL,
  sender_process,
  sender_postprocess,
  sender_uninit,
  sender_methods
};

typedef struct _MSPlcDetector {
  uint32_t prev_packet_ssrc;
  uint32_t prev_packet_ts;
  uint16_t prev_packet_cseq;
  uint32_t diff_packet_interval_ts;
  uint32_t prev_msrtp_ts;
  uint32_t max_plc_frame;
}MSPlcDetector;

struct _ReceiverData {
  RtpSession *session;
  int rate;
  int errs;
  int max_inbuffer;
  int rtp_burst;
  int reset;
  bool_t mute;
  MSPlcDetector ms_plc;
  int rtp_received;
  int rtp_loss1;
  int rtp_loss2;
  int rtp_loss3;
  uint16_t rtp_first_cseq;
  uint16_t rtp_last_cseq;
};

typedef struct _ReceiverData ReceiverData;

static void receiver_init(MSFilter * f)
{
  ReceiverData *d = (ReceiverData *)ms_new0(ReceiverData, 1);
  d->session = NULL;
  d->rate = 8000;
  d->errs = 0;
  d->max_inbuffer = 0;
  d->rtp_burst = 0;
  d->mute = FALSE;
  f->data = d;
}

static void receiver_postprocess(MSFilter * f){
  ReceiverData *d = (ReceiverData *) f->data;
  if (d->session!=NULL)
    rtp_session_disable_sendrecv(d->session);
}

static void receiver_uninit(MSFilter * f){
  ReceiverData *d = (ReceiverData *) f->data;
  ms_warning("msrtp: incoming RTP burst detected during conversation: %i", d->rtp_burst);
  ms_message("msrtp: RTP packet cseq=%i->%i rtp=%i plc1=%i plc2=%i plc3=%i", d->rtp_first_cseq, d->rtp_last_cseq, d->rtp_received, d->rtp_loss1, d->rtp_loss2, d->rtp_loss3);
  ms_free(d);
}

static int receiver_set_session(MSFilter * f, void *arg)
{
  ReceiverData *d = (ReceiverData *) f->data;
  RtpSession *s = (RtpSession *) arg;
  PayloadType *pt = rtp_profile_get_payload(rtp_session_get_recv_profile(s),
    rtp_session_get_recv_payload_type
    (s));
  int oldrate=d->rate;
  if (pt != NULL) {
    if (strcasecmp("g722", pt->mime_type)==0 )
      d->rate=8000;
    else d->rate = pt->clock_rate;
  } else {
    ms_warning("Receiving undefined payload type %i ?",
      rtp_session_get_recv_payload_type(s));
  }
  if (d->session==NULL)
    d->session = s;
  else {
    d->session = s;
    if (oldrate!=d->rate)
      d->reset = 1;
  }

  return 0;
}

static int receiver_get_sr(MSFilter *f, void *arg){
  ReceiverData *d = (ReceiverData *) f->data;
  PayloadType *pt;
  if (d->session==NULL) {
    ms_warning("Could not obtain sample rate, session is not set.");
    return -1;
  }
  pt=rtp_profile_get_payload(rtp_session_get_recv_profile(d->session),
    rtp_session_get_recv_payload_type(d->session));
  if (pt != NULL) {
    *(int*)arg=pt->clock_rate;
  }else{
    ms_warning("Could not obtain sample rate, payload type is unknown.");
    return -1;
  }
  return 0;
}

static void receiver_preprocess(MSFilter * f){
  ReceiverData *d = (ReceiverData *) f->data;
  d->errs = 0;
  d->max_inbuffer = 0;
  d->rtp_burst = 0;
  memset(&d->ms_plc, 0, sizeof(MSPlcDetector));
  if (d->session){
    PayloadType *pt=rtp_profile_get_payload(
      rtp_session_get_recv_profile(d->session),
      rtp_session_get_recv_payload_type(d->session));
    if (pt){
      if (pt->type!=PAYLOAD_VIDEO)
        rtp_session_flush_sockets(d->session);
    }
  }
}

static mblk_t *packet_exist(RtpSession *session, uint16_t cseq, int *cseq_offset) {
  rtp_header_t *tmprtp;
  mblk_t *tmp;
  queue_t *q = &session->rtp.rq;
  *cseq_offset=0;
  if (qempty(q))
  {
    return NULL;
  }
  for(tmp=qbegin(q);!qend(q,tmp);tmp=qnext(q,tmp)){
    tmprtp=(rtp_header_t*)tmp->b_rptr;
    if (tmprtp->seq_number==cseq)
    {
      *cseq_offset=0;
      return tmp;
    }
    if (tmprtp->seq_number==cseq+1)
    {
      *cseq_offset=1;
      return tmp;
    }
    if (tmprtp->seq_number==cseq+2)
    {
      *cseq_offset=2;
      return tmp;
    }
    if (tmprtp->seq_number==cseq+3)
    {
      *cseq_offset=3;
      return tmp;
    }
  }
  return NULL;
}

static void receiver_packet_found(MSFilter * f, uint32_t timestamp, mblk_t *m, int check_loss)
{
  ReceiverData *d = (ReceiverData *) f->data;

  if (d->rtp_first_cseq==0)
    d->rtp_first_cseq = rtp_get_seqnumber(m);
  d->rtp_last_cseq = rtp_get_seqnumber(m);
  d->rtp_received++;

  mblk_set_timestamp_info(m, rtp_get_timestamp(m));
  mblk_set_marker_info(m, rtp_get_markbit(m));
  mblk_set_payload_type(m, rtp_get_payload_type(m));
  mblk_set_cseq(m, rtp_get_seqnumber(m));

  if (d->ms_plc.prev_packet_ssrc != rtp_get_ssrc(m)) {
    ms_message("%s: first RTP packet -> ssrc=%X timestamp = %u // cseq = %u", f->ticker->name, rtp_get_ssrc(m), rtp_get_timestamp(m), rtp_get_seqnumber(m));
    d->ms_plc.prev_packet_ssrc = rtp_get_ssrc(m);
    d->ms_plc.diff_packet_interval_ts = 0;
  } else if (rtp_get_markbit(m)==1) {
    d->ms_plc.diff_packet_interval_ts = 0;
  } else if (d->ms_plc.prev_packet_ts==0) {
    d->ms_plc.diff_packet_interval_ts = 0;
  } else {
    int cseq_offset = rtp_get_seqnumber(m)-d->ms_plc.prev_packet_cseq;
    if (cseq_offset==1 && rtp_get_timestamp(m)>d->ms_plc.prev_packet_ts && rtp_get_timestamp(m)-d->ms_plc.prev_packet_ts<=(uint32_t)(d->rate/50)*7) {
      d->ms_plc.diff_packet_interval_ts = rtp_get_timestamp(m)-d->ms_plc.prev_packet_ts;
      if (d->session->rtcpxr_voip_metrics) {
        if (d->session->rtp.stats.frameduration_ms==0)
          ms_message("%s: frameduration_ms: %i", f->ticker->name, (d->ms_plc.diff_packet_interval_ts)*1000/d->rate);
        d->session->rtp.stats.frameduration_ms = (d->ms_plc.diff_packet_interval_ts)*1000/d->rate;
      }
    }
    if (cseq_offset>1 && rtp_get_timestamp(m)>d->ms_plc.prev_packet_ts && rtp_get_timestamp(m)-d->ms_plc.prev_packet_ts<=(uint32_t)(d->rate/50)*7) {
      PayloadType *pt=rtp_profile_get_payload(rtp_session_get_profile(d->session), rtp_session_get_recv_payload_type(d->session));
      if (pt && pt->type!=PAYLOAD_VIDEO) {
        mblk_t *empty;

        if (check_loss>0) {
          d->rtp_loss1++;
          if ((strcasecmp("silk", pt->mime_type)==0 || strcasecmp("opus", pt->mime_type)==0)) {
            /* need PLC: insert empty block? */
            empty = dupb(m);
            rtp_get_payload(empty, &empty->b_rptr);
            mblk_set_plc(empty, cseq_offset-1);
          } else {
            empty = allocb(0, 0);
            mblk_set_plc(empty, 0x1);
          }
#ifndef DISABLE_PLC_LOG
          ms_message("non continuous cseq detected: PLC required (%i) (packet interval=%i), prev_cseq=%i, cur_cseq=%i, diff_ts=%i queuesize=%i", cseq_offset-1,
            d->ms_plc.diff_packet_interval_ts, d->ms_plc.prev_packet_cseq, rtp_get_seqnumber(m), timestamp-d->ms_plc.prev_msrtp_ts, d->session->rtp.rq.q_mcount);
#endif
          ms_queue_put(f->outputs[0], empty);
        } else {
#ifndef DISABLE_PLC_LOG
          ms_message("non continuous cseq detected: no PLC for late packets (%i) (packet interval=%i), prev_cseq=%i, cur_cseq=%i, diff_ts=%i queuesize=%i", cseq_offset-1,
            d->ms_plc.diff_packet_interval_ts, d->ms_plc.prev_packet_cseq, rtp_get_seqnumber(m), timestamp-d->ms_plc.prev_msrtp_ts, d->session->rtp.rq.q_mcount);
#endif
#if 0
		  /* we just had 2 late packet (at least). If we do not add an empty packets, we may got a tick. */
          /* unfortunatly, this is pushing more packet than expected in the msconf filter, */
          /* however, the msconf has internal workaround to handle a growing queue. */
          if ((strcasecmp("silk", pt->mime_type) == 0 || strcasecmp("opus", pt->mime_type) == 0)) {
            /* need PLC: insert empty block? */
            empty = dupb(m);
            rtp_get_payload(empty, &empty->b_rptr);
            mblk_set_plc(empty, cseq_offset - 1);
          }
          else {
            empty = allocb(0, 0);
            mblk_set_plc(empty, 0x1);
          }
          ms_queue_put(f->outputs[0], empty);
#endif
        }
      }

    }
  }
  d->ms_plc.prev_msrtp_ts = timestamp;
  d->ms_plc.prev_packet_ts = rtp_get_timestamp(m);
  d->ms_plc.prev_packet_cseq = rtp_get_seqnumber(m);
  d->ms_plc.max_plc_frame = 5;

  rtp_get_payload(m,&m->b_rptr);
  if (d->mute != TRUE)
    ms_queue_put(f->outputs[0], m);
  else
    freemsg(m);
  if (d->max_inbuffer<d->session->rtp.rq.q_mcount)
    d->max_inbuffer = d->session->rtp.rq.q_mcount;
  if (rtp_session_get_last_recv_error_code(d->session)>0)
  {
    rtp_session_clear_recv_error_code(d->session);
    d->errs++;
  }
}

static void receiver_process(MSFilter * f)
{
  ReceiverData *d = (ReceiverData *) f->data;
  mblk_t *m;
  uint32_t timestamp;
  mblk_t *last_dropped=NULL;

  if (d->session == NULL)
    return;

  if (d->session->bundle!=NULL) rtp_bundle_lock(d->session->bundle);
  rtp_session_enable_sendrecv (d->session);

  if (d->reset == 1) {
    d->reset=0;
    d->ms_plc.prev_packet_ssrc=0;
  }

  timestamp = (uint32_t) (f->ticker->time * (d->rate/1000));
  while ((m = rtp_session_recvm_with_ts_extended(d->session, timestamp, &last_dropped)) != NULL) {

    if (d->reset == 1) {
      d->reset=0;
      timestamp = (uint32_t) (f->ticker->time * (d->rate/1000));
      d->ms_plc.prev_packet_ssrc=0;
    }

    if (last_dropped!=NULL) {
      if (d->ms_plc.prev_packet_cseq>=rtp_get_seqnumber(last_dropped)) {
#ifndef DISABLE_PLC_LOG
        ms_message("late packet discarded: already rebuilt cseq=%i, cur_cseq=%i, queuesize=%i", 
          d->ms_plc.prev_packet_cseq, rtp_get_seqnumber(last_dropped), d->session->rtp.rq.q_mcount);
#endif
        freemsg(last_dropped);
      } else {
#ifndef DISABLE_PLC_LOG
        //ms_message("late packet detected: PLC required cseq=%i, cur_cseq=%i, queuesize=%i", 
        //  d->ms_plc.prev_packet_cseq, rtp_get_seqnumber(last_dropped), d->session->rtp.rq.q_mcount);
#endif
        receiver_packet_found(f, timestamp, last_dropped, 0);
      }
    }
    receiver_packet_found(f, timestamp, m, 1);
  }

  if (d->session->rtp.jittctl.enabled==TRUE && !d->session->permissive){
    PayloadType *pt=rtp_profile_get_payload(rtp_session_get_profile(d->session), rtp_session_get_recv_payload_type(d->session));
    if (pt && pt->type!=PAYLOAD_VIDEO) {
      if (d->ms_plc.diff_packet_interval_ts>0 && d->ms_plc.diff_packet_interval_ts<=timestamp-d->ms_plc.prev_msrtp_ts) {
        int cseq_offset=0;
        mblk_t *empty;
        mblk_t *res = packet_exist(d->session, d->ms_plc.prev_packet_cseq+1, &cseq_offset);
        if (res!=NULL && cseq_offset>0) {
          d->rtp_loss2++;
          if ((strcasecmp("silk", pt->mime_type)==0 || strcasecmp("opus", pt->mime_type)==0) && cseq_offset>0) {
            /* need PLC: insert empty block? */
            empty = dupb(res);
            rtp_get_payload(empty, &empty->b_rptr);
            mblk_set_plc(empty, cseq_offset);
          } else {
            empty = allocb(0, 0);
            mblk_set_plc(empty, 0x1);
          }
#ifndef DISABLE_PLC_LOG
          ms_message("PLC required/packet loss detection (%i) (packet interval=%i), prev_cseq=%i, diff_ts=%i queuesize=%i", cseq_offset,
            d->ms_plc.diff_packet_interval_ts, d->ms_plc.prev_packet_cseq, timestamp-d->ms_plc.prev_msrtp_ts, d->session->rtp.rq.q_mcount);
#endif
          /* update gap and burst voip-metrics */
          rtp_session_update_voip_metrics(d->session, 0, 1, 0);

          ms_queue_put(f->outputs[0], empty);

          /* act as if a real packet has been sent: */
          d->ms_plc.prev_msrtp_ts = timestamp;
          d->ms_plc.prev_packet_ts += d->ms_plc.diff_packet_interval_ts;
          d->ms_plc.prev_packet_cseq++;
          d->ms_plc.max_plc_frame--;

        } else if (d->session->rtp.rq.q_mcount==0 && d->ms_plc.max_plc_frame>0) {
          d->rtp_loss3++;
          empty = allocb(0, 0);
          mblk_set_plc(empty, 0x1);
#ifndef DISABLE_PLC_LOG
          ms_message("PLC required/packet loss detection/empty queue (%i) (packet interval=%i), prev_cseq=%i, diff_ts=%i", cseq_offset,
            d->ms_plc.diff_packet_interval_ts, d->ms_plc.prev_packet_cseq, timestamp-d->ms_plc.prev_msrtp_ts);
#endif
          ms_queue_put(f->outputs[0], empty);

          /* act as if a real packet has been sent: */
          d->ms_plc.prev_msrtp_ts = timestamp;
          d->ms_plc.prev_packet_ts += d->ms_plc.diff_packet_interval_ts;
          d->ms_plc.prev_packet_cseq++;
          d->ms_plc.max_plc_frame--;
        }
      }
    }
  }
  if (d->session->bundle!=NULL) rtp_bundle_unlock(d->session->bundle);

  if (f->outputs[0]->q.q_mcount>5)
  {
    //ms_warning("RTP burst detected: %i", f->outputs[0]->q.q_mcount);
    d->rtp_burst++;
  }
}


static int receiver_get_statistics(MSFilter *f, void *arg){
  ReceiverData *d = (ReceiverData *) f->data;
  ms_warning("filter: %s[%i->%i]", f->desc->name,
    f->desc->ninputs, f->desc->noutputs);
  if (d->session!=NULL)
  {
    ms_warning("filter: %s RTP=%i packets -> out[0]=%i", f->desc->name, 
      d->session->rtp.rq.q_mcount,
      (f->outputs[0]!=NULL)?f->outputs[0]->q.q_mcount:-1);
  }
  ms_warning("filter: %s max_inbuffer=%i packets // sendto errs=%i", f->desc->name, 
    d->max_inbuffer,
    d->errs);

  return 0;
}

static int receiver_mute_session(MSFilter *f, void *arg){
  ReceiverData *d = (ReceiverData *) f->data;
 
  if (d->session == NULL)
    return 0;

  d->mute = *((int*)arg);

  return 0;
}

static MSFilterMethod receiver_methods[] = {
  {	MS_RTP_RECV_SET_SESSION	, receiver_set_session	},
  {	MS_FILTER_GET_SAMPLE_RATE	, receiver_get_sr		},
  {	MS_FILTER_GET_STATISTICS, receiver_get_statistics },
  {	MS_RTP_RECV_MUTE_SESSION, receiver_mute_session },
  {	0, NULL}
};

MSFilterDesc ms_rtp_recv_desc = {
  MS_RTP_RECV_ID,
  "MSRtpRecv",
  N_("RTP input filter"),
  MS_FILTER_OTHER,
  NULL,
  0,
  1,
  receiver_init,
  receiver_preprocess,
  receiver_process,
  receiver_postprocess,
  receiver_uninit,
  receiver_methods
};

MS_FILTER_DESC_EXPORT(ms_rtp_send_desc)
MS_FILTER_DESC_EXPORT(ms_rtp_recv_desc)
