/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include "mediastreamer-config.h"
#endif

#ifdef __APPLE__
#include "TargetConditionals.h"
#endif

#if (_MSC_VER >= 1700) && !defined(_USING_V110_SDK71_)
#include <winapifamily.h>
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#define DISABLE_PRIORITY
#include <iostream>
#include <chrono>
#include <thread>
#endif
#endif

#include "mediastreamer2/msticker.h"

#if defined(__linux)
#include <inttypes.h>
#endif
#if defined(HAVE_SCHED_H) && defined(_GNU_SOURCE)
#include <sched.h>
#endif

#if defined (WIN32)
#define MEDIASTREAMER_PERFORMANCE
#elif defined(__linux)
#define MEDIASTREAMER_PERFORMANCE
#if defined(__ARM_ARCH__) && (__ARM_ARCH__ < 7)
#undef MEDIASTREAMER_PERFORMANCE
#endif
#endif

static int get_cpu() {
#ifdef WIN32
  SYSTEM_INFO sysinfo;
  GetSystemInfo(&sysinfo);

  return sysinfo.dwNumberOfProcessors;
#elif defined(ANDROID)
  return sysconf(_SC_NPROCESSORS_ONLN);
#elif (TARGET_OS_IPHONE == 1)
  return 1;
#elif defined(__APPLE__) && defined(__OBJC__)
  return (int)[[NSProcessInfo processInfo] activeProcessorCount];
#elif defined(__linux)
  return sysconf(_SC_NPROCESSORS_ONLN);
#else
  return 0;
#endif
}

void * ms_ticker_run(void *s);
void * ms_ticker_run_core(void *arg);
static uint64_t get_cur_time(void *);

#ifdef MEDIASTREAMER_PERFORMANCE
static uint64_t get_perf_time();

static void ms_perf_init(MSFilter *f){
  memset(&f->usage, 0, sizeof(MSPerf));
}

static void ms_perf_start(MSFilter *f){
  f->usage.prev_start = get_perf_time();
}

static void ms_perf_stop(MSFilter *f){
  uint64_t stop = get_perf_time();
  if (stop<f->usage.prev_start)
    return;
  f->usage.total += (stop-f->usage.prev_start);
  if (stop-f->usage.prev_start>f->ticker->interval*1000)
    f->usage.late_counter++;
  if (stop-f->usage.prev_start>100000)
    ms_message("perf issue %s %i:", f->desc->name, (int)(stop-f->usage.prev_start));
}

static void ms_perf_log(MSFilter *f){
#if defined(__linux)
  ms_message("perf for %s : %" PRIu64 "", f->desc->name, f->usage.total);
#elif !defined(WIN32)
  ms_message("perf for %s : %lld", f->desc->name, f->usage.total);
#else
	LARGE_INTEGER m_freq;
	QueryPerformanceFrequency(&m_freq);
  ms_message("perf for %s : %I64d", f->desc->name, f->usage.total);
#endif
  if (f->usage.late_counter>0)
    ms_message("perf for %s : late counter = %i", f->desc->name, f->usage.late_counter);
}
#endif


static void ms_ticker_start(MSTicker *s) {
  int idx;
  int max_core = MIN(s->to_core - s->from_core + 1, 32);
  for (idx = 0; idx < max_core; idx++) {
    s->tasks[idx].idx = s->from_core + idx;
    s->tasks[idx].run = TRUE;
    s->tasks[idx].ticker = s;
    ms_mutex_init(&s->tasks[idx].thread_lock_run, NULL);
    ms_cond_init(&s->tasks[idx].thread_cond_run, NULL);
    ms_mutex_init(&s->tasks[idx].thread_lock_done, NULL);
    ms_cond_init(&s->tasks[idx].thread_cond_done, NULL);

    ms_thread_create(&s->tasks[idx].thread, NULL, ms_ticker_run_core, &s->tasks[idx]);
    {
      char tmp[16];
      snprintf(tmp, sizeof(tmp), "%.11s-t%d", s->name, s->tasks[idx].idx);
      tmp[15] = '\0';
      ms_thread_set_name(s->tasks[idx].thread, tmp);
    }
  }
  s->run = TRUE;
  ms_thread_create(&s->thread, NULL, ms_ticker_run, s);
  {
    char tmp[16];
    snprintf(tmp, sizeof(tmp), "%s", s->name);
    tmp[15] = '\0';
    ms_thread_set_name(s->thread, tmp);
  }
}

static void ms_ticker_init(MSTicker *ticker, const char *name)
{
	ms_mutex_init(&ticker->lock,NULL);
	ticker->execution_list=NULL;
	ticker->ticks=1;
	ticker->time=0;
	ticker->interval=20;
  ticker->total_cores = get_cpu();
  if (ticker->total_cores < 1)
    ticker->total_cores = 1;
#if defined(ANDROID)
	if (strstr(name, "amsip-audio") != NULL)
	  ticker->interval=10;
#endif
	if (strstr(name, "amsip-video") != NULL)
	  ticker->interval=30;
	if (strstr(name, "amsip-screensharing") != NULL)
	  ticker->interval=80;
	if (strstr(name, "amsip-screensharing-rcv") != NULL)
	  ticker->interval=40;

  ticker->from_core = 0;
  ticker->to_core = ticker->total_cores - 1;
  if (strstr(name, "from-core") != NULL) {

    char value[12];
    ticker->from_core = (ticker->total_cores<4)?0:(ticker->total_cores-4);
    ticker->to_core = ticker->total_cores - 1;
    if (fmtp_get_value(name, "from-core", value, sizeof(value))) {
      ticker->from_core = atoi(value);
      ms_message("from-core: %d", ticker->from_core);
    }
    if (fmtp_get_value(name, "to-core", value, sizeof(value))) {
      ticker->to_core = atoi(value);
      ms_message("to-core: %d", ticker->to_core);
    }
  }
  else if (strstr(name, "audio") != NULL) {
    ticker->from_core = 0;
    ticker->to_core = (ticker->total_cores<4) ? 0 : 1;
  }
  else if (strstr(name, "video") != NULL) {
    ticker->from_core = (ticker->total_cores<4) ? 1 : 2;
    if (ticker->from_core + 1 > ticker->total_cores) {
      ticker->from_core = ticker->total_cores - 1;
      ticker->to_core = ticker->total_cores - 1;
    }
    else {
      ticker->to_core = ticker->total_cores - 1;
    }
  }

	ticker->run=FALSE;
	ticker->exec_id=0;
	ticker->get_cur_time_ptr=&get_cur_time;
	ticker->get_cur_time_data=NULL;
	ticker->name=ms_strdup(name);

  {
    char *tmp = strstr(ticker->name, ";");
    if (tmp != NULL) {
      tmp[0] = '\0';
    }
  }
  ms_ticker_start(ticker);
}

MSTicker *ms_ticker_new(){
	MSTicker *obj=(MSTicker *)ms_new0(MSTicker,1);
	ms_ticker_init(obj, "MSTicker");
	return obj;
}

MSTicker *ms_ticker_new_withname(const char *name){
	MSTicker *obj=(MSTicker *)ms_new0(MSTicker,1);
	ms_ticker_init(obj, name);
	return obj;
}

static void ms_ticker_stop(MSTicker *s){
  int idx;
  int max_core = MIN(s->to_core - s->from_core + 1, 32);
  ms_mutex_lock(&s->lock);

  for (idx = 0; idx < max_core; idx++) {
    if (s->tasks[idx].thread == 0)
      break;
    ms_mutex_lock(&s->tasks[idx].thread_lock_run);
    s->tasks[idx].run = FALSE;
    ms_cond_signal(&s->tasks[idx].thread_cond_run);
    ms_mutex_unlock(&s->tasks[idx].thread_lock_run);
    ms_thread_join(s->tasks[idx].thread, NULL);
    ms_mutex_destroy(&s->tasks[idx].thread_lock_run);
    ms_cond_destroy(&s->tasks[idx].thread_cond_run);
    ms_mutex_destroy(&s->tasks[idx].thread_lock_done);
    ms_cond_destroy(&s->tasks[idx].thread_cond_done);
  }

	s->run=FALSE;
	ms_mutex_unlock(&s->lock);
	ms_thread_join(s->thread,NULL);
}

void ms_ticker_set_name(MSTicker *s, const char *name){
	if (s->name) ms_free(s->name);
	s->name=ms_strdup(name);
}

static void ms_ticker_uninit(MSTicker *ticker)
{
	ms_ticker_stop(ticker);
	ms_free(ticker->name);
	ms_mutex_destroy(&ticker->lock);
}

void ms_ticker_destroy(MSTicker *ticker){
	ms_ticker_uninit(ticker);
	ms_free(ticker);
}


static MSList *get_sources(MSList *filters){
	MSList *sources=NULL;
	MSFilter *f;
	for(;filters!=NULL;filters=filters->next){
		f=(MSFilter*)filters->data;
		if (f->desc->ninputs==0){
			sources=ms_list_append(sources,f);
		}
	}
	return sources;
}

int ms_ticker_attach(MSTicker *ticker,MSFilter *f)
{
	MSList *sources=NULL;
	MSList *filters=NULL;
	MSList *it;
	
	if (f->ticker!=NULL) {
		ms_message("Filter %s is already being scheduled; nothing to do.",f->desc->name);
		return 0;
	}

	filters=ms_filter_find_neighbours(f);
	sources=get_sources(filters);
	if (sources==NULL){
		ms_fatal("No sources found around filter %s",f->desc->name);
		ms_list_free(filters);
		return -1;
	}
	/*run preprocess on each filter: */
	for(it=filters;it!=NULL;it=it->next)
	{
#ifdef MEDIASTREAMER_PERFORMANCE
	  ms_perf_init((MSFilter*)it->data);
#endif
	  ms_filter_preprocess((MSFilter*)it->data,ticker);
	}
	ms_mutex_lock(&ticker->lock);
	ticker->execution_list = ms_list_concat(ticker->execution_list,sources);
  ms_mutex_unlock(&ticker->lock);
	ms_list_free(filters);
	return 0;
}

int ms_ticker_detach(MSTicker *ticker,MSFilter *f){
	MSList *sources=NULL;
	MSList *filters=NULL;
	MSList *it;

	if (f->ticker==NULL) {
		ms_message("Filter %s is not scheduled; nothing to do.",f->desc->name);
		return 0;
	}

	ms_mutex_lock(&ticker->lock);

	filters=ms_filter_find_neighbours(f);
	sources=get_sources(filters);
	if (sources==NULL){
		ms_fatal("No sources found around filter %s",f->desc->name);
		ms_list_free(filters);
		ms_mutex_unlock(&ticker->lock);
		return -1;
	}

	for(it=sources;it!=NULL;it=ms_list_next(it)){
		ticker->execution_list=ms_list_remove(ticker->execution_list,it->data);
	}
	ms_mutex_unlock(&ticker->lock);
	ms_list_for_each(filters,(void (*)(void*))ms_filter_postprocess);
#ifdef MEDIASTREAMER_PERFORMANCE
	ms_list_for_each(filters,(void (*)(void*))ms_perf_log);
#endif

	ms_list_free(filters);
	ms_list_free(sources);
	return 0;
}


static bool_t filter_can_process(MSFilter *f, int tick){
	/* look if filters before this one have run */
	int i;
	MSQueue *l;
	for(i=0;i<f->desc->ninputs;i++){
		l=f->inputs[i];
		if (l!=NULL){
			if (l->prev.filter->last_tick!=tick) return FALSE;
		}
	}
	return TRUE;
}

static void call_process(MSFilter *f){
	bool_t process_done=FALSE;
	if (f->desc->ninputs==0 || f->desc->flags & MS_FILTER_IS_PUMP){
#ifdef MEDIASTREAMER_PERFORMANCE
		ms_perf_start(f);
#endif
		ms_filter_process(f);
#ifdef MEDIASTREAMER_PERFORMANCE
		ms_perf_stop(f);
#endif
	}else{
		while (ms_filter_inputs_have_data(f)) {
			if (process_done){
				ms_warning("Re-scheduling filter %s: all data should be consumed in one process call, so fix it.",f->desc->name);
			}
#ifdef MEDIASTREAMER_PERFORMANCE
			ms_perf_start(f);
#endif
			ms_filter_process(f);
#ifdef MEDIASTREAMER_PERFORMANCE
			ms_perf_stop(f);
#endif
			process_done=TRUE;
		}
	}
}

static void run_graph(MSFilter *f, MSTicker *s, MSList **unschedulable, bool_t force_schedule){
	int i;
	MSQueue *l;
	if (f->last_tick!=s->ticks ){
		if (filter_can_process(f,s->ticks) || force_schedule) {
			/* this is a candidate */
			f->last_tick=s->ticks;
      call_process(f);
			/* now recurse to next filters */		
			for(i=0;i<f->desc->noutputs;i++){
				l=f->outputs[i];
				if (l!=NULL){
          run_graph(l->next.filter,s,unschedulable, force_schedule);
				}
			}
		}else{
			/* this filter has not all inputs that have been filled by filters before it. */
			*unschedulable=ms_list_prepend(*unschedulable,f);
		}
	}
}

static void run_graphs(MSTicker *s, MSList *execution_list, bool_t force_schedule){
	MSList *it;
	MSList *unschedulable=NULL;
	for(it=execution_list;it!=NULL;it=it->next){
		run_graph((MSFilter*)it->data,s,&unschedulable,force_schedule);
	}
	/* filters that are part of a loop haven't been called in process() because one of their input refers to a filter that could not be scheduled (because they could not be scheduled themselves)... Do you understand ?*/
	/* we resolve this by simply assuming that they must be called anyway 
	for the loop to run correctly*/
	/* we just recall run_graphs on them, as if they were source filters */
	if (unschedulable!=NULL) {
		run_graphs(s,unschedulable,TRUE);
		ms_list_free(unschedulable);
	}
}

static void run_graphs2(MSTicker *s, MSList *execution_list, bool_t force_schedule) {

  MSList *unschedulable = NULL;
  MSList *it;
  int n_core = 0;
  int max_core = MIN(s->to_core - s->from_core + 1, 32);
  if (execution_list == NULL)
    return;

  if (max_core == 0) {
    run_graphs(s, execution_list, force_schedule);
    return;
  }

  /* build execution list for each thread */
  for (it = execution_list; it != NULL; it = it->next) {
    MSFilter *filter = (MSFilter *)it->data;
    s->tasks[n_core%max_core].execution_list = ms_list_append(s->tasks[n_core%max_core].execution_list, filter);
    n_core++;
  }

  /* wakeup all thread */
  for (n_core = 0; n_core < max_core; n_core++) {
    ms_mutex_lock(&s->tasks[n_core].thread_lock_run);
    s->tasks[n_core].action_done = FALSE;
    ms_cond_signal(&s->tasks[n_core].thread_cond_run); //on windows, sometimes, this line is called before ms_cond_wait?
    ms_mutex_unlock(&s->tasks[n_core].thread_lock_run);
  }

  /* wait for all thread */
  for (n_core = 0; n_core < max_core; n_core++) {
    ms_mutex_lock(&s->tasks[n_core].thread_lock_done);
#ifdef WIN32
    /* On windows, every previous ms_cond_signal must have a ms_cond_wait after (different implementation) */
    ms_cond_wait(&s->tasks[n_core].thread_cond_done, &s->tasks[n_core].thread_lock_done);
#else
    if (!s->tasks[n_core].action_done)
      ms_cond_wait(&s->tasks[n_core].thread_cond_done, &s->tasks[n_core].thread_lock_done);
#endif
    ms_mutex_unlock(&s->tasks[n_core].thread_lock_done);
    ms_list_free(s->tasks[n_core].execution_list);
    s->tasks[n_core].execution_list = NULL;
  }

  /* now recurse to next filters */
  for (it = execution_list; it != NULL; it = it->next) {
    int i;
    MSFilter *f = (MSFilter *)it->data;
    for (i = 0; i < f->desc->noutputs; i++) {
      MSQueue *l;
      l = f->outputs[i];
      if (l != NULL) {
        run_graph(l->next.filter, s, &unschedulable, force_schedule);
      }
    }
  }
  if (unschedulable != NULL) {
    run_graphs(s, unschedulable, TRUE);
    ms_list_free(unschedulable);
  }
}

#ifdef __MACH__
#include <mach/clock.h>
#include <mach/mach.h>
#include <sys/types.h>
#include <sys/timeb.h>
#endif

static uint64_t get_cur_time(void *unused){
#if defined(_WIN32_WCE)
	DWORD timemillis = GetTickCount();
	return timemillis;
#elif defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#if 1
	std::chrono::monotonic_clock::time_point val = std::chrono::monotonic_clock::now();
	std::chrono::milliseconds ms = std::chrono::duration_cast<std::chrono::milliseconds>(val.time_since_epoch());
	return ms.count();
#else
	return GetTickCount64();
#endif
#elif defined(WIN32)
	static uint32_t rollover_count=0;
	static DWORD last_t=0;
	DWORD t = timeGetTime();

	if ( last_t > t && (last_t - t > 100000) )
	{
		/* timeGetTime() may returns lower time than before because
			we are receiving values from different processors.
		*/
		ms_warning("get_cur_time: rollover detected last_t = %u", last_t);
		ms_warning("get_cur_time: rollover detected cur_t  = %u", t);
		++rollover_count;
	}
	last_t = t;
	return ((uint64_t)rollover_count<<32) + t;
#elif defined(__MACH__) && defined(__GNUC__) && (__GNUC__ >= 3)
  clock_serv_t cclock;
  mach_timespec_t mts;
  host_get_clock_service (mach_host_self (), SYSTEM_CLOCK, &cclock);
  clock_get_time (cclock, &mts);
  mach_port_deallocate (mach_task_self (), cclock);
  return (mts.tv_sec*1000LL) + (mts.tv_nsec/1000000LL);
#elif defined(__MACH__)
	struct timespec ts;
	struct timeb time_val;
	
	ftime (&time_val);
	ts.tv_sec = time_val.time;
	ts.tv_nsec = time_val.millitm * 1000000;
	return (ts.tv_sec*1000LL) + (ts.tv_nsec/1000000LL);
#else
	struct timespec ts;
	if (clock_gettime(CLOCK_MONOTONIC,&ts)<0){
		ms_fatal("clock_gettime() doesn't work: %s",ortp_strerror(errno));
	}
	return (ts.tv_sec*1000LL) + (ts.tv_nsec/1000000LL);
#endif
}

#ifdef MEDIASTREAMER_PERFORMANCE
static uint64_t get_perf_time(){
#if defined(_WIN32_WCE)
	DWORD timemillis = GetTickCount();
	return timemillis*1000;
#elif defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#if 1
	std::chrono::monotonic_clock::time_point val = std::chrono::monotonic_clock::now();
	std::chrono::milliseconds ms = std::chrono::duration_cast<std::chrono::milliseconds>(val.time_since_epoch());
	return ms.count()*1000;
#else
	return GetTickCount64()*1000;
#endif
#elif defined(WIN32)
  return (uint64_t)(timeGetTime()*1000);
#elif defined(__MACH__) && defined(__GNUC__) && (__GNUC__ >= 3)
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (tv.tv_sec*1000000LL) + (tv.tv_usec*1LL);
#elif defined(__MACH__)
	struct timespec ts;
	struct timeb time_val;
	
	ftime (&time_val);
	ts.tv_sec = time_val.time;
	ts.tv_nsec = time_val.millitm * 1000000;
	return (ts.tv_sec*1000000LL) + (time_val.millitm*1000LL);
#else
	struct timespec ts;
	if (clock_gettime(CLOCK_MONOTONIC,&ts)<0){
		ms_fatal("clock_gettime() doesn't work: %s",ortp_strerror(errno));
	}
	return (ts.tv_sec*1000LL) + (ts.tv_nsec/1000000LL);
#endif
}
#endif


static void sleepMs(int ms){
#if defined(WINAPI_FAMILY) && (WINAPI_FAMILY == WINAPI_FAMILY_PHONE_APP)
#if 1
  std::chrono::milliseconds dura( ms );
  std::this_thread::sleep_for( dura );
#else
  struct timeval delay;
  delay.tv_sec = 0;
  delay.tv_usec = ms*1000;
  select (0, 0, 0, 0, &delay);
#endif
#elif defined(WIN32)
	Sleep(ms);
#else
	struct timespec ts;
	ts.tv_sec=0;
	ts.tv_nsec=ms*1000000LL;
	nanosleep(&ts,NULL);
#endif
}

static int set_high_prio(void){
	int precision=2;
	int result=0;
#ifdef DISABLE_PRIORITY
#elif defined(WIN32)
	MMRESULT mm;
	TIMECAPS ptc;
	mm=timeGetDevCaps(&ptc,sizeof(ptc));
	if (mm==0){
		if (ptc.wPeriodMin<(UINT)precision)
			ptc.wPeriodMin=precision;
		else
			precision = ptc.wPeriodMin;
		mm=timeBeginPeriod(ptc.wPeriodMin);
		if (mm!=TIMERR_NOERROR){
			ms_warning("timeBeginPeriod failed.");
		}
		ms_message("win32 timer resolution set to %i ms",ptc.wPeriodMin);
	}else{
		ms_warning("timeGetDevCaps failed.");
	}
  
	if(!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL)){
		ms_warning("SetThreadPriority(THREAD_PRIORITY_TIME_CRITICAL) failed (%d)\n", GetLastError());
		if(!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST)){
			ms_warning("SetThreadPriority(THREAD_PRIORITY_HIGHEST) failed (%d)\n", GetLastError());
		}
	}
#elif defined(ANDROID)
	setpriority(PRIO_PROCESS, 0, -20); //higher priority
#else
	struct sched_param param;
	memset(&param,0,sizeof(param));
	int policy=SCHED_RR;
	param.sched_priority=sched_get_priority_max(policy);
	if((result=pthread_setschedparam(pthread_self(),policy, &param))) {
		ms_warning("no real time priority access (err=%i)",result);
		policy=SCHED_OTHER;
		param.sched_priority=sched_get_priority_max(policy);
		if((result=pthread_setschedparam(pthread_self(),policy, &param))) {
		  ms_warning("failed to increase priority (err=%i)",result);
		}
	}
	if (result==0) {
		ms_message("MS ticker priority set to max");
	}
#endif
	return precision;
}

static void unset_high_prio(int precision){
#ifdef DISABLE_PRIORITY
#elif defined(WIN32)
	MMRESULT mm;

	if(!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL)){
		ms_warning("SetThreadPriority() failed (%d)\n", GetLastError());
	}

	mm=timeEndPeriod(precision);
#endif
}

#if defined(HAVE_SCHED_H) && defined(_GNU_SOURCE)
int stick_this_thread_to_core(int from_core, int to_core) {
  pthread_t current_thread;
  cpu_set_t cpuset;
  int num_cores = sysconf(_SC_NPROCESSORS_ONLN);
  int j;
  if (num_cores <= 1)
    return 0;
  if (to_core >= num_cores)
    return EINVAL;

  CPU_ZERO(&cpuset);
  for (j = from_core; j <= to_core; j++) {
    CPU_SET(j, &cpuset);
    ms_message("enable core: %d", j);
  }
  current_thread = pthread_self();
  return pthread_setaffinity_np(current_thread, sizeof(cpu_set_t), &cpuset);
}
#endif

void * ms_ticker_run_core(void *arg)
{
  MSTickerTask *tt = (MSTickerTask*)arg;
  MSTicker *s = tt->ticker;

#if defined(HAVE_SCHED_H) && defined(_GNU_SOURCE)
  stick_this_thread_to_core(tt->idx, tt->idx);
#endif

  tt->action_done=TRUE;
  while (tt->run) {
    MSList *it;

    ms_mutex_lock(&tt->thread_lock_run);
    if (tt->run) {
#ifdef WIN32
      ms_cond_wait(&tt->thread_cond_run, &tt->thread_lock_run);
#else
      if (tt->action_done) {
        ms_cond_wait(&tt->thread_cond_run, &tt->thread_lock_run);
      }
#endif
    }
    ms_mutex_unlock(&tt->thread_lock_run);

    if (!tt->run) {
      break;
    }
    for (it = tt->execution_list; it != NULL; it = it->next) {
      MSFilter* f = (MSFilter*)it->data;
      if (f->last_tick != s->ticks) {
        f->last_tick = s->ticks;
        call_process(f);
      }
    }
    ms_mutex_lock(&tt->thread_lock_done);
    tt->action_done = TRUE;
    ms_cond_signal(&tt->thread_cond_done);
    ms_mutex_unlock(&tt->thread_lock_done);
  }
  ms_message("%s ms_ticker_run_core[%i] exiting", s->name, tt->idx);
  ms_thread_exit(NULL);
  return NULL;
}

void * ms_ticker_run(void *arg)
{
  uint64_t realtime;
  int64_t diff;
  MSTicker *s=(MSTicker*)arg;
  int lastlate=0;
  int precision=2;
  int late;
  bool_t skip_late=FALSE;

  if (strstr(s->name, "amsip-video")!=NULL)
    skip_late=TRUE;

#if TARGET_OS_IPHONE
  if (strstr(s->name, "audio") != NULL)
    precision = set_high_prio();
#else
  precision = set_high_prio();

#if defined(HAVE_SCHED_H) && defined(_GNU_SOURCE)
  stick_this_thread_to_core(s->from_core, s->to_core);
#endif

#endif

  s->ticks=1;
  ms_mutex_lock(&s->lock);
  s->orig=s->get_cur_time_ptr(s->get_cur_time_data);

  while(s->run){
    s->ticks++;
    if (s->ticks==200)
      ms_ticker_print_graphs(s);
    run_graphs2(s,s->execution_list,FALSE);
    s->time+=s->interval;
    while(1){
      realtime=s->get_cur_time_ptr(s->get_cur_time_data)-s->orig;
      ms_mutex_unlock(&s->lock);
      diff=s->time-realtime;
      if (strstr(s->name, "noadjust") != NULL) {
        /* allow OS scheduler to switch to other threads */
        diff = 0;
      }
      if (diff>0){
        /* sleep until next tick */
        sleepMs((int)diff);
      }else{
        /* allow OS scheduler to switch to other threads */
        sleepMs(1);
        late=(int)-diff;
        if (late>s->interval*5 && late>lastlate){
          if (skip_late==TRUE)
          {
            ms_warning("%s: skip filter processing to adjust.(+%i)",s->name,late-(late%s->interval));
            s->time+=late-(late%s->interval);
          } else {
            //ms_warning("%s: We are late of %d miliseconds.",s->name,late);
          }
        } else if (late>s->interval*50*30 && strstr(s->name, "noadjust") == NULL) {
          ms_warning("%s: adjusting time -most probably after hibernate-.(+%i)",s->name,late-(late%s->interval));
          s->time+=late-(late%s->interval);
        }
        lastlate=late;
        break; /*exit the while loop */
      }
      ms_mutex_lock(&s->lock);
    }
    ms_mutex_lock(&s->lock);
  }
  ms_mutex_unlock(&s->lock);

#if TARGET_OS_IPHONE
  if (strstr(s->name, "audio") != NULL)
    unset_high_prio(precision);
#else
  unset_high_prio(precision);
#endif

  ms_message("%s thread exiting",s->name);

  ms_thread_exit(NULL);
  return NULL;
}

void ms_ticker_set_time_func(MSTicker *ticker, MSTickerTimeFunc func, void *user_data){
	if (func==NULL) func=get_cur_time;
	/*ms_mutex_lock(&ticker->lock);*/
	ticker->get_cur_time_ptr=func;
	ticker->get_cur_time_data=user_data;
	/*re-set the origin to take in account that previous function ptr and the
	new one may return different times*/
	ticker->orig=func(user_data)-ticker->time;
	/*ms_mutex_unlock(&ticker->lock);*/
	ms_message("ms_ticker_set_time_func: ticker updated.");
}

static void print_graph(MSFilter *f, MSTicker *s, MSList **unschedulable, bool_t force_schedule){
	int i;
	MSQueue *l;
	if (f->last_tick!=s->ticks ){
		if (filter_can_process(f,s->ticks) || force_schedule) {
			/* this is a candidate */
			f->last_tick=s->ticks;
      ms_message("print_graphs: %s", f->desc->name);
			/* now recurse to next filters */		
			for(i=0;i<f->desc->noutputs;i++){
				l=f->outputs[i];
				if (l!=NULL){
					print_graph(l->next.filter,s,unschedulable, force_schedule);
				}
			}
		}else{
			/* this filter has not all inputs that have been filled by filters before it. */
			*unschedulable=ms_list_prepend(*unschedulable,f);
		}
	}
}

static void print_graphs(MSTicker *s, MSList *execution_list, bool_t force_schedule){
	MSList *it;
	MSList *unschedulable=NULL;
	for(it=execution_list;it!=NULL;it=it->next){
		print_graph((MSFilter*)it->data,s,&unschedulable,force_schedule);
	}
	/* filters that are part of a loop haven't been called in process() because one of their input refers to a filter that could not be scheduled (because they could not be scheduled themselves)... Do you understand ?*/
	/* we resolve this by simply assuming that they must be called anyway 
	for the loop to run correctly*/
	/* we just recall run_graphs on them, as if they were source filters */
	if (unschedulable!=NULL) {
		print_graphs(s,unschedulable,TRUE);
		ms_list_free(unschedulable);
	}
}

void ms_ticker_print_graphs(MSTicker *ticker){
	print_graphs(ticker,ticker->execution_list,FALSE);
}
