/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include "mediastreamer-config.h"
#endif

#ifdef HAVE_LINUX_VIDEODEV2_H

#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <poll.h>

#include <linux/videodev2.h>

#include "mediastreamer2/msvideo.h"
#include "mediastreamer2/msticker.h"
#include "mediastreamer2/mswebcam.h"

#ifdef HAVE_LIBV4L2
#include <libv4l2.h>
#else

#define v4l2_open open
#define v4l2_close close
#define v4l2_mmap mmap
#define v4l2_munmap munmap
#define v4l2_ioctl ioctl

#endif

#ifdef HAVE_LINUX_UVCVIDEO_H

#ifndef V4L2_PIX_FMT_H264
#define V4L2_PIX_FMT_H264 v4l2_fourcc('H', '2', '6', '4')
#endif
#include <linux/uvcvideo.h>
#include <linux/usb/video.h>
#include "uvc_h264.h"
#endif

#ifdef HAVE_LIBUSB_H
#include <libusb.h>
#endif

typedef struct V4l2State{
	int fd;
	char *dev;
	char *mmapdbuf;
	int msize;/*mmapped size*/
	MSVideoSize vsize;
	MSVideoSize got_vsize;
	int pix_fmt;
	int int_pix_fmt; /*internal pixel format */
	int picture_size;
	mblk_t *frames[VIDEO_MAX_FRAME];
	int frame_ind;
	int frame_max;
	float fps;
	float start_time;
	int frame_count;
	int queued;
	bool_t configured;
#ifdef HAVE_LINUX_UVCVIDEO_H
  uint8_t unit;
#endif
  uint32_t bitrate;
  uint32_t max_bitrate;
  int send_idr;
  uint32_t reset;
}V4l2State;


#ifdef HAVE_LINUX_UVCVIDEO_H

#ifdef HAVE_LIBUSB_H
typedef struct
{
  int8_t bLength;
  int8_t bDescriptorType;
  int8_t bDescriptorSubType;
  int8_t bUnitID;
  uint8_t guidExtensionCode[16];
} __attribute__ ((__packed__)) xu_descriptor;

#define USB_VIDEO_CONTROL 0x01
#define USB_VIDEO_CONTROL_INTERFACE 0x24
#define USB_VIDEO_CONTROL_XU_TYPE 0x06

static uint8_t _libusb_get_unit_id_dev(libusb_device *device) {

  struct libusb_device_descriptor desc;
  int i,j,k;
  static const uint8_t guid[16] = GUID_UVCX_H264_XU;

  if (libusb_get_device_descriptor (device, &desc) == 0) {
    for (i = 0; i < desc.bNumConfigurations; ++i) {
      struct libusb_config_descriptor *config = NULL;
      if (libusb_get_config_descriptor (device, i, &config) == 0) {
	for (j = 0; j < config->bNumInterfaces; j++) {
	  for (k = 0; k < config->interface[j].num_altsetting; k++) {
	    const struct libusb_interface_descriptor *interface;
	    const uint8_t *ptr = NULL;
	    interface = &config->interface[j].altsetting[k];
	    if (interface->bInterfaceClass != LIBUSB_CLASS_VIDEO ||
		interface->bInterfaceSubClass != USB_VIDEO_CONTROL)
	      continue;
	    ptr = interface->extra;
	    while (ptr - interface->extra + sizeof (xu_descriptor) < interface->extra_length) {
	      xu_descriptor *xu_desc = (xu_descriptor *) ptr;
	      if (xu_desc->bDescriptorType == USB_VIDEO_CONTROL_INTERFACE &&
		  xu_desc->bDescriptorSubType == USB_VIDEO_CONTROL_XU_TYPE &&
		  memcmp (xu_desc->guidExtensionCode, guid, 16) == 0) {
		uint8_t unit_id = xu_desc->bUnitID;
		ms_message("Device %x:%x Class: %i || unit_id for H264=%i", desc.idVendor, desc.idProduct, (int)desc.bDeviceClass, unit_id);
		return unit_id;
	      }
	      ptr += xu_desc->bLength;
	    }
	  }
	}
	libusb_free_config_descriptor(config);
      }
    }
  }
  return 0;
}

static uint8_t _libusb_get_unit_id() {
  /* use libusb */
  libusb_context *usb_ctx = NULL;
  libusb_device **device_list = NULL;
  libusb_device *device = NULL;
  ssize_t cnt;
  int i;

  uint8_t unit_id=0;

  if (usb_ctx == NULL)
    libusb_init (&usb_ctx);

  cnt = libusb_get_device_list (usb_ctx, &device_list);
  for (i = 0; i < cnt; i++)
    {
      device = libusb_ref_device (device_list[i]);
      unit_id = _libusb_get_unit_id_dev(device);
      libusb_unref_device (device);
      if (unit_id)
	break;
    }

  libusb_free_device_list (device_list, 1);
  return unit_id;
}

#endif

struct __uvc_xu_find_unit
{
  uint8_t guid[16];
  uint8_t unit;
} __attribute__((packed)) ;

#define __UVCIOC_XU_FIND_UNIT _IOWR('u', 0x22, struct __uvc_xu_find_unit)

static int find_unit(V4l2State *s) {
  static const __u8 guid[16] = GUID_UVCX_H264_XU;
  struct __uvc_xu_find_unit xu;

#ifdef HAVE_LIBUSB_H
  return _libusb_get_unit_id();
#endif
  
  memcpy (xu.guid, guid, 16);
  xu.unit = 0;

  if (v4l2_ioctl (s->fd, __UVCIOC_XU_FIND_UNIT, &xu)<0) {
    ms_error("__UVCIOC_XU_FIND_UNIT: %i:%s",errno, strerror(errno));
    return 12;
  }
  return xu.unit;
}

static int uvc_ioctl(V4l2State *s, uint32_t selector, uint32_t query, uint8_t *data) {
  struct uvc_xu_control_query xu;
  uint16_t len;

  if (s->unit==0)
    {
      s->unit=find_unit(s);
      ms_message("xu unit: %i", s->unit);
    }
  xu.unit = s->unit;

  xu.selector = selector;
  xu.query = UVC_GET_LEN;
  xu.size = sizeof(len);
  xu.data = (unsigned char *)&len;

  if (v4l2_ioctl (s->fd, UVCIOC_CTRL_QUERY, &xu)<0){
    ms_error("UVC_GET_LEN: %i:%s",errno, strerror(errno));
    return -1;
  }
  
  if (query == UVC_GET_LEN) {
    *((uint16_t *) data) = len;
  } else {
    xu.query = query;
    xu.size = len;
    xu.data = data;
    if (v4l2_ioctl (s->fd, UVCIOC_CTRL_QUERY, &xu)<0) {
      return -1;
    }
  }

  return 0;
}

#define MSV4L2_DEBUG
#ifdef MSV4L2_DEBUG
static void
print_probe_commit (uvcx_video_config_probe_commit_t * probe)
{
  ms_message("  Frame interval : %d *100ns",
		    probe->dwFrameInterval);
  ms_message("  Bit rate : %d", probe->dwBitRate);
  ms_message("  Hints : %X", probe->bmHints);
  ms_message("  Configuration index : %d",
		    probe->wConfigurationIndex);
  ms_message("  Width : %d", probe->wWidth);
  ms_message("  Height : %d", probe->wHeight);
  ms_message("  Slice units : %d", probe->wSliceUnits);
  ms_message("  Slice mode : %X", probe->wSliceMode);
  ms_message("  Profile : %X", probe->wProfile);
  ms_message("  IFrame Period : %d ms", probe->wIFramePeriod);
  ms_message("  Estimated video delay : %d ms",
		    probe->wEstimatedVideoDelay);
  ms_message("  Estimated max config delay : %d ms",
		    probe->wEstimatedMaxConfigDelay);
  ms_message("  Usage type : %X", probe->bUsageType);
  ms_message("  Rate control mode : %X", probe->bRateControlMode);
  ms_message("  Temporal scale mode : %X",
		    probe->bTemporalScaleMode);
  ms_message("  Spatial scale mode : %X",
		    probe->bSpatialScaleMode);
  ms_message("  SNR scale mode : %X", probe->bSNRScaleMode);
  ms_message("  Stream mux option : %X", probe->bStreamMuxOption);
  ms_message("  Stream Format : %X", probe->bStreamFormat);
  ms_message("  Entropy CABAC : %X", probe->bEntropyCABAC);
  ms_message("  Timestamp : %X", probe->bTimestamp);
  ms_message("  Num of reorder frames : %d",
		    probe->bNumOfReorderFrames);
  ms_message("  Preview flipped : %X", probe->bPreviewFlipped);
  ms_message("  View : %d", probe->bView);
  ms_message("  Stream ID : %X", probe->bStreamID);
  ms_message("  Spatial layer ratio : %f",
		    ((probe->bSpatialLayerRatio & 0xF0) >> 4) +
		    ((float) (probe->bSpatialLayerRatio & 0x0F)) / 16);
  ms_message("  Leaky bucket size : %d ms",
		    probe->wLeakyBucketSize);
}
#endif

static int test_h264(V4l2State *s) {
  {
    uvcx_encoder_reset req = { 0 };
    if (uvc_ioctl (s, UVCX_ENCODER_RESET, UVC_SET_CUR, (uint8_t *) & req)<0) {
      ms_error("UVCX_ENCODER_RESET SET_CUR %s" ,strerror(errno));
      s->unit=0;
      return -1;
    }
  }
  return 0;
}

static int configure_h264(V4l2State *s) {
  uvcx_video_config_probe_commit_t probe;

  {
    uvcx_encoder_reset req = { 0 };
    if (uvc_ioctl (s, UVCX_ENCODER_RESET, UVC_SET_CUR, (uint8_t *) & req)<0) {
      ms_error("UVCX_ENCODER_RESET SET_CUR %s" ,strerror(errno));
      return -1;
    }
  }

  memset (&probe, 0, sizeof (probe));
  probe.dwFrameInterval = 1000000; //333333;

  probe.wWidth = 320;
  probe.wHeight = 240;
  probe.bStreamMuxOption = 5;

  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_PROBE, UVC_SET_CUR, (uint8_t *) & probe)<0) {
    ms_error("uvc_ioctl PROBE UVC_SET_CUR %s" ,strerror(errno));
    return -1;
  }

  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_PROBE, UVC_GET_CUR, (uint8_t *) & probe)<0) {
    ms_error("uvc_ioctl PROBE UVC_GET_CUR %s" ,strerror(errno));
    return -1;
  }

#ifdef MSV4L2_DEBUG
  ms_message("PROBE UVC_SET_CUR : ");
  print_probe_commit (&probe);
#endif

  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_COMMIT, UVC_SET_CUR, (uint8_t *) & probe)<0) {
    ms_error("uvc_ioctl COMMIT UVC_SET_CUR %s" ,strerror(errno));
    return -1;
  }


  /* Print MIN/MAX/DEF probe values for debugging purposes */
  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_PROBE, UVC_GET_MIN, (uint8_t *) & probe)<0) {
    return -1;
  }
#ifdef MSV4L2_DEBUG
  ms_message("PROBE GET_MIN : ");
  print_probe_commit (&probe);
#endif

  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_PROBE, UVC_GET_MAX, (uint8_t *) & probe)<0) {
    ms_error("uvc_ioctl UVC_GET_MIN %s" ,strerror(errno));
    return -1;
  }

#ifdef MSV4L2_DEBUG
  ms_message("PROBE GET_MAX : ");
  print_probe_commit (&probe);
#endif

  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_PROBE, UVC_GET_DEF, (uint8_t *) & probe)<0) {
    ms_error("uvc_ioctl UVC_GET_MIN %s" ,strerror(errno));
    return -1;
  }

#ifdef MSV4L2_DEBUG
  ms_message("PROBE GET_DEF : ");
  print_probe_commit (&probe);
#endif

  probe.dwFrameInterval = 1000000; /* 333333; */
  probe.dwBitRate=s->bitrate;
  if (probe.dwBitRate>800000) probe.dwBitRate=800000; /* current limit */
  probe.wProfile=0x4240; /* baseline // 40 */
  probe.wWidth=s->vsize.width;
  probe.wHeight=s->vsize.height;
  probe.wIFramePeriod=5000; /* 5 sec */
  probe.bRateControlMode = UVC_H264_RATECONTROL_CBR;
  probe.bUsageType = UVC_H264_USAGETYPE_REALTIME; //BROADCAST;
  probe.bmHints = 0xFFFF;
  probe.bStreamMuxOption=3;
  //probe.wSliceUnits=8;
  //probe.wSliceMode=3;
  if (s->bitrate<800000) {
    probe.wSliceUnits=6000;
    probe.wSliceMode=1;
  } else {
    probe.wSliceUnits=4;
    probe.wSliceMode=3;
  }


#ifdef MSV4L2_DEBUG
  ms_message("PROBE TRY : ");
  print_probe_commit (&probe);
#endif

  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_PROBE, UVC_SET_CUR, (uint8_t *) & probe)<0) {
    ms_error("uvc_ioctl UVC_GET_MIN %s" ,strerror(errno));
    return -1;
  }

  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_PROBE, UVC_GET_CUR, (uint8_t *) & probe)<0) {
    ms_error("uvc_ioctl UVC_GET_MIN %s" ,strerror(errno));
    return -1;
  }

#ifdef MSV4L2_DEBUG
  ms_message("PROBE GET_CUR : ");
  print_probe_commit (&probe);
#endif

  if (uvc_ioctl (s, UVCX_VIDEO_CONFIG_COMMIT, UVC_SET_CUR, (uint8_t *) & probe)<0) {
    ms_error("uvc_ioctl UVC_GET_MIN %s" ,strerror(errno));
    return -1;
  }
  s->vsize.width=probe.wWidth;
  s->vsize.height=probe.wHeight;
  return 0;
}

static int update_h264_settings(V4l2State *s) {

  struct __bitrate {
    uint16_t id;
    uint32_t peak;
    uint32_t average;
    } __attribute__((packed)) ;
  
  struct __level_idc
  {
    uint16_t id;
    uint32_t mbmax;
    uint8_t level_idc;
    uint8_t reserved;
  } __attribute__((packed));
  
  struct __bitrate val;
#if 0
  struct __level_idc val_level;
#endif
  
  {
    uvcx_encoder_reset req = { 0 };
    if (uvc_ioctl (s, UVCX_ENCODER_RESET, UVC_SET_CUR, (uint8_t *) & req)<0) {
      ms_error("UVCX_ENCODER_RESET SET_CUR %s" ,strerror(errno));
    }
  }
  
#if 0
  if (uvc_ioctl(s, 0x0D, UVC_GET_CUR, (uint8_t*) &val_level)==0) {
    ms_message("val_level.level_idc: %u", val_level.level_idc);
    ms_message("val_level.mbmax: %u", val_level.mbmax);
    val_level.level_idc = 12;
    val_level.mbmax = 1024000;
    if (uvc_ioctl(s, 0x0D, UVC_SET_CUR, (uint8_t*) &val_level)<0) {
      ms_error("failed to set bitrate");
    }
    
    if (uvc_ioctl(s, 0x0D, UVC_GET_CUR, (uint8_t*) &val_level)==0) {
      ms_message("val_level.level_idc: %u", val_level.level_idc);
      ms_message("val_level.mbmax: %u", val_level.mbmax);
    }
  }
#endif
  
  if (uvc_ioctl(s, 0x0E, UVC_GET_CUR, (uint8_t*) &val)==0) {
    ms_message("val.peak: %u", val.peak);
    ms_message("val.average: %u", val.average);
    val.peak = s->bitrate;
    val.average = s->bitrate;
    if (uvc_ioctl(s, 0x0E, UVC_SET_CUR, (uint8_t*) &val)<0) {
      ms_error("failed to set bitrate");
    }
    
    if (uvc_ioctl(s, 0x0E, UVC_GET_CUR, (uint8_t*) &val)==0) {
      ms_message("val.peak: %u", val.peak);
      ms_message("val.average: %u", val.average);
    }
  }

  return 0;
}

static void save_sps_pps(MSFilter *f, mblk_t *im){
  uint8_t type = (*im->b_rptr) & ((1<<5)-1);

  if (type==7) {
    ms_message("new SPS provided");
  } else if (type==8) {
    ms_message("new PPS provided");
  }
}

static void msv4l2_h264_handle(MSFilter *f, mblk_t *im){
  V4l2State *s=(V4l2State*)f->data;
  uint32_t ts=(uint32_t)(f->ticker->time*90LL);
  uint8_t *beg=NULL;
  uint8_t *src=im->b_rptr;

  mblk_t *om = dupb(im);

  while (im->b_wptr-src>4) {

    if (src[0]==0 && src[1]==0 && src[2]==3) {
      /* skip preventing bytes */
      src=src+4; /* skip 003 and next bytes (0?) */
      continue;
    }

    if (src[0]==0 && src[1]==0 && src[2]==1) {
      if (beg!=NULL) {
	mblk_t *out;
	om->b_rptr = beg;
	om->b_wptr = src-1;

	out=copyb(om);
	mblk_set_timestamp_info(out, ts);
	mblk_set_video_format(out,s->pix_fmt);
	if (s->bitrate>512000)
	  mblk_set_precious_flag(out, 1);
	//if (msgdsize(out)>1100)
	//ms_message("size partial NAL: %i", msgdsize(out));
	save_sps_pps(f, out);
	ms_queue_put(f->outputs[0],out);
      }
      src=src+3;
      beg=src;
      continue;
    } else if (src[0]==0 && src[1]==0 && src[2]==0 && src[3]==0) {
      if (beg!=NULL) {
	mblk_t *out;
	om->b_rptr = beg;
	om->b_wptr = src; //-1;

	out=copyb(om);
	mblk_set_timestamp_info(out, ts);
	mblk_set_video_format(out,s->pix_fmt);
	if (s->bitrate>512000)
	  mblk_set_precious_flag(out, 1);
	//if (msgdsize(out)>1100)
	//ms_message("size partial NAL: %i", msgdsize(out));
	save_sps_pps(f, out);
	ms_queue_put(f->outputs[0],out);
      }
      return;
    }
    
    src++;
  }

  if (beg!=NULL) {
    mblk_t *out;

    //static int discard=0;
    //discard++;
    //if (discard%10==0) {
      //freemsg(om);
      //return;
    //}
    om->b_rptr = beg;
    om->b_wptr = im->b_wptr;
    
    out=copyb(om);
    mblk_set_timestamp_info(out, ts);
    mblk_set_video_format(out,s->pix_fmt);
    mblk_set_marker_info(out,1); /* last NAL */
    if (s->bitrate>512000)
      mblk_set_precious_flag(out, 1);
    //if (msgdsize(out)>1100)
    //ms_message("size final NAL: %i", msgdsize(out));
    save_sps_pps(f, out);
    ms_queue_put(f->outputs[0],out);
  }

  freemsg(om);
}

#endif

static int msv4l2_open(V4l2State *s){
	int fd=v4l2_open(s->dev,O_RDWR|O_NONBLOCK);
	if (fd==-1){
		ms_error("Could not open %s: %s",s->dev,strerror(errno));
		return -1;
	}
	s->fd=fd;
	return 0;
}

static int msv4l2_close(V4l2State *s){
	if (s->fd!=-1){
		v4l2_close(s->fd);
		s->fd=-1;
		s->configured=FALSE;
	}
	return 0;
}

static int v4lv2_try_format( V4l2State *s, struct v4l2_format *fmt, int fmtid){
	
	fmt->type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt->fmt.pix.pixelformat = fmtid;
	fmt->fmt.pix.field = V4L2_FIELD_ANY;

        if (v4l2_ioctl (s->fd, VIDIOC_TRY_FMT, fmt)<0){
	  ms_message("VIDIOC_TRY_FMT: %i:%s",errno,strerror(errno));
	  return errno;
	}
	if (fmt->fmt.pix.pixelformat != fmtid) return -1;
	if (v4l2_ioctl (s->fd, VIDIOC_S_FMT, fmt)<0){
	  ms_message("VIDIOC_S_FMT: %i:%s",errno, strerror(errno));
	  if (errno!=16)
	    return errno;
	}

	return 0;
}

static int get_picture_buffer_size(MSPixFmt pix_fmt, int w, int h){
	switch(pix_fmt){
		case MS_YUV420P:
			return (w*h*3)/2;
		break;
		case MS_RGB24:
			return w*h*3;
		break;
		case MS_YUYV:
			return w*h*2;
		break;
		default:
			return 0;
	}
	return 0;
}

static int msv4l2_configure(V4l2State *s){
	struct v4l2_capability cap;
	struct v4l2_format fmt;
	struct v4l2_streamparm params;
	MSVideoSize vsize;
	
        if (v4l2_ioctl (s->fd, VIDIOC_QUERYCAP, &cap)<0) {
		ms_message("Not a v4lv2 driver.");
		return -1;
        }

        if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
		ms_error("%s is not a video capture device\n",s->dev);
		return -1;
        }

	if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
		ms_error("%s does not support streaming i/o\n",s->dev);
		return -1;
	}
	
	ms_message("Driver is %s // card is %s // bus_info is %s ",cap.driver,cap.card, cap.bus_info);
	memset(&fmt,0,sizeof(fmt));

	fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (v4l2_ioctl (s->fd, VIDIOC_G_FMT, &fmt)<0){
		ms_error("VIDIOC_G_FMT failed: %s",strerror(errno));
	}
	vsize=s->vsize;

#ifdef HAVE_LINUX_UVCVIDEO_H
	if (ms_filter_codec_supported("H264")) {

	  ms_message("msv4l2_configure: size & bitrate being applied (br=%i size=%ix%i)", s->bitrate, s->vsize.width, s->vsize.height);
	  if (s->vsize.width>=MS_VIDEO_SIZE_720P_W && s->bitrate<=512000)
	    {
	      s->vsize.width=MS_VIDEO_SIZE_CIF_W;
	      s->vsize.height=MS_VIDEO_SIZE_CIF_H;
	      ms_message("msv4l2_configure: vsize has been modified to CIF because bandwidth is low");
	    }

	  if (test_h264(s)==0)
	    s->pix_fmt=MS_H264;
	  if (s->unit!=0 && s->pix_fmt==MS_H264) {
	    int i;
	    //fmt.fmt.pix.width       = 320;
	    //fmt.fmt.pix.height      = 240;
	    fmt.fmt.pix.width       = s->vsize.width;
	    fmt.fmt.pix.height      = s->vsize.height;
	    s->pix_fmt=0;
	    i = v4lv2_try_format(s,&fmt,V4L2_PIX_FMT_H264);
	    if (i==0){
	      s->int_pix_fmt=V4L2_PIX_FMT_H264;
	      ms_message("v4lv2: H264 choosen");
	      if (configure_h264(s)==0)
		{
		  s->pix_fmt=MS_H264;
		  ms_message("Size of webcam delivered pictures is %ix%i",s->vsize.width,s->vsize.height);
		}
	    }
	  }
	}
#endif

	if (s->unit==0) {
	  do{
	    int i;
	    fmt.fmt.pix.width       = s->vsize.width;
	    fmt.fmt.pix.height      = s->vsize.height;

#ifdef ENABLE_H264_WITHOUT_UVC
	    i = v4lv2_try_format(s,&fmt,V4L2_PIX_FMT_H264);
	    if (i==5)
	      break;
	    if (i==0){
	      s->pix_fmt=MS_H264;
	      s->int_pix_fmt=V4L2_PIX_FMT_H264;
	      ms_message("v4lv2: H264 choosen (no uvcvideo)");
	      break;
	    }
#endif
	    i = v4lv2_try_format(s,&fmt,V4L2_PIX_FMT_YUV420);
	    if (i==5)
	      break;
	    if (i==0){
	      s->pix_fmt=MS_YUV420P;
	      s->int_pix_fmt=V4L2_PIX_FMT_YUV420;
	      ms_message("v4lv2: YUV420P choosen");
	      break;
	    }
	    i = v4lv2_try_format(s,&fmt,V4L2_PIX_FMT_YUYV);
	    if (i==5)
	      break;
	    if (i==0){
	      s->pix_fmt=MS_YUYV;
	      s->int_pix_fmt=V4L2_PIX_FMT_YUYV;
	      ms_message("v4lv2: V4L2_PIX_FMT_YUYV choosen");
	      break;
	    }
	    i = v4lv2_try_format(s,&fmt,V4L2_PIX_FMT_RGB24);
	    if (i==5)
	      break;
	    if (i==0){
	      s->pix_fmt=MS_RGB24;
	      s->int_pix_fmt=V4L2_PIX_FMT_RGB24;
	      ms_message("v4lv2: RGB24 choosen");
	      break;
	    }
	    i = v4lv2_try_format(s,&fmt,V4L2_PIX_FMT_MJPEG);
	    if (i==5)
	      break;
	    if (i==0){
	      s->pix_fmt=MS_MJPEG;
	      s->int_pix_fmt=V4L2_PIX_FMT_MJPEG;
	      ms_message("v4lv2: MJPEG choosen");
	      break;
	    }
	    
	    ms_error("Could not find supported pixel format for %ix%i", s->vsize.width, s->vsize.height);
	    s->vsize=ms_video_size_get_just_lower_than(s->vsize);
	  }while(s->vsize.width!=0);

	  if (s->vsize.width==0){
	    ms_message("Could not find any combination of resolution/pixel-format that works !");
	    s->vsize=vsize;
	    return -1;
	  }
	}

	memset(&fmt,0,sizeof(fmt));

	fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (v4l2_ioctl (s->fd, VIDIOC_G_FMT, &fmt)<0){
		ms_error("VIDIOC_G_FMT failed: %s",strerror(errno));
	}else{
		ms_message("Size of webcam delivered pictures is %ix%i",fmt.fmt.pix.width,fmt.fmt.pix.height);
		s->vsize.width=fmt.fmt.pix.width;
		s->vsize.height=fmt.fmt.pix.height;
	}

	memset(&params, 0, sizeof(params));
	params.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	if (v4l2_ioctl (s->fd, VIDIOC_G_PARM, &fmt)<0){
	  ms_warning("VIDIOC_G_PARM: %i:%s",errno, strerror(errno));
	}else{
	  params.parm.capture.timeperframe.numerator = 1;
	  params.parm.capture.timeperframe.denominator = (uint32_t)15;
	  if (v4l2_ioctl (s->fd, VIDIOC_S_PARM, &params)<0){
	    ms_warning("VIDIOC_S_PARM failed: %s",strerror(errno));
	  }
	}
	
	s->picture_size=get_picture_buffer_size(s->pix_fmt,s->vsize.width,s->vsize.height);
	s->configured=TRUE;

	return 0;
}

static int msv4l2_do_mmap(V4l2State *s){
	struct v4l2_requestbuffers req;
	int i;
	enum v4l2_buf_type type;
	
	memset(&req,0,sizeof(req));
	
	if (s->pix_fmt==MS_H264)
	  req.count               = 1; /* uv4l driver on raspberry is providing SPS/PPS for all buffer + no IDR... */
	else
	  req.count               = 4;

	req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory              = V4L2_MEMORY_MMAP;
	
	if (v4l2_ioctl (s->fd, VIDIOC_REQBUFS, &req)<0) {
		ms_error("Error requesting info on mmap'd buffers: %s",strerror(errno));
		return -1;
	}
	
	for (i=0; i<req.count; ++i) {
		struct v4l2_buffer buf;
		mblk_t *msg;
		void *start;
		memset(&buf,0,sizeof(buf));
	
		buf.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory=V4L2_MEMORY_MMAP;
		buf.index=i;
	
		if (v4l2_ioctl (s->fd, VIDIOC_QUERYBUF, &buf)<0){
			ms_error("Could not VIDIOC_QUERYBUF : %s",strerror(errno));
			return -1;
		}
		
		start=v4l2_mmap (NULL /* start anywhere */,
			buf.length,
			PROT_READ | PROT_WRITE /* required */,
			MAP_SHARED /* recommended */,
			s->fd, buf.m.offset);
	
		if (start==NULL){
			ms_error("Could not v4l2_mmap: %s",strerror(errno));
		}
		msg=esballoc(start,buf.length,0,NULL);
		msg->b_wptr+=buf.length;
		s->frames[i]=msg;
	}
	s->frame_max=req.count;
	for (i = 0; i < s->frame_max; ++i) {
		struct v4l2_buffer buf;

		memset(&buf,0,sizeof(buf));
		buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory      = V4L2_MEMORY_MMAP;
		buf.index       = i;
		if (-1==v4l2_ioctl (s->fd, VIDIOC_QBUF, &buf)){
			ms_error("VIDIOC_QBUF failed: %s",strerror(errno));
		}else {
			s->frames[i]->b_datap->db_ref++;
			s->queued++;
		}
	}

	/*start capture immediately*/
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (-1 ==v4l2_ioctl (s->fd, VIDIOC_STREAMON, &type)){
		ms_error("VIDIOC_STREAMON failed: %s",strerror(errno));
		return -1;
	}

#ifdef HAVE_LINUX_UVCVIDEO_H
	if (s->pix_fmt==MS_H264)
	  update_h264_settings(s);
#endif

#define V4L2_CID_MPEG_VIDEO_REPEAT_SEQ_HEADER           (V4L2_CID_MPEG_BASE+226)

	if (s->pix_fmt==MS_H264 && s->unit==0) {
	  struct v4l2_control ctrl;
	  ctrl.id = V4L2_CID_MPEG_VIDEO_REPEAT_SEQ_HEADER;
	  ctrl.value = 1;
	  if (-1 ==v4l2_ioctl (s->fd, VIDIOC_S_CTRL, &ctrl)) {
	    ms_error("V4L2_CID_MPEG_VIDEO_REPEAT_SEQ_HEADER failed: %s",strerror(errno));
	  } else
	    ms_message("V4L2_CID_MPEG_VIDEO_REPEAT_SEQ_HEADER done");
	}

	return 0;
}

static void
set_bitrate (V4l2State *s, uint32_t bandwidth)
{
  uvcx_bitrate_layers_t req;

  if (s->pix_fmt==MS_H264 && s->unit==0) {
    if (uvc_ioctl (s, UVCX_BITRATE_LAYERS, UVC_GET_CUR, (uint8_t *) & req)<0) {
      ms_warning("UVCX_BITRATE_LAYERS GET_CUR %s" ,strerror(errno));
    }
    
    req.dwPeakBitrate = bandwidth;
    req.dwAverageBitrate = bandwidth;
    if (uvc_ioctl (s, UVCX_BITRATE_LAYERS, UVC_SET_CUR, (uint8_t *) & req)<0) {
      ms_warning("UVCX_BITRATE_LAYERS SET_CUR %s" ,strerror(errno));
    }
  } else {
    s->reset=1;
  }
}

static mblk_t * v4lv2_grab_image(V4l2State *s, bool_t keyframe){
	struct v4l2_buffer buf;
	unsigned int k;
	memset(&buf,0,sizeof(buf));
	mblk_t *ret=NULL;

	buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf.memory = V4L2_MEMORY_MMAP;

	if (keyframe==TRUE) {
	  if (s->pix_fmt==MS_H264) {
	    
#ifdef HAVE_LINUX_UVCVIDEO_H
	    if (s->unit!=0) {
	      /* this does work for my logitech C920 camera! */
	      uvcx_picture_type_control_t req = { 0, 0 };
	      req.wPicType = UVC_H264_PICTYPE_IDR_WITH_PPS_SPS;
	      if (uvc_ioctl (s, UVCX_PICTURE_TYPE_CONTROL, UVC_SET_CUR, (uint8_t *) & req)<0) {
		ms_warning("UVCX_PICTURE_TYPE_CONTROL SET_CUR %s" ,strerror(errno));
	      }
	    }

#if 0
	    struct v4l2_control ctrl;
#define V4L2_CID_MPEG_VIDEO_FORCE_KEY_FRAME       (V4L2_CID_MPEG_BASE+229)
	    /* both doesn't work */
	    ctrl.id = V4L2_CID_MPEG_VIDEO_FORCE_KEY_FRAME;
	    ctrl.id = V4L2_CID_MPEG_MFC51_VIDEO_FORCE_FRAME_TYPE;
	    ctrl.value = 1;
	    if (-1 ==v4l2_ioctl (s->fd, VIDIOC_S_CTRL, &ctrl)) {
	      ms_error("V4L2_CID_MPEG_VIDEO_FORCE_KEY_FRAME failed: %s",strerror(errno));
	    } else
	      ms_message("V4L2_CID_MPEG_VIDEO_FORCE_KEY_FRAME done");
#endif
	    
#endif
	    /* this doesn't work for my logitech C920 camera... */
	    buf.flags = V4L2_BUF_FLAG_KEYFRAME;
	  }
	}
	
	/*queue buffers whose ref count is 1, because they are not
	still used anywhere in the filter chain */
	for(k=0;k<s->frame_max;++k){
		if (s->frames[k]->b_datap->db_ref==1){
			buf.index=k;
			if (-1==v4l2_ioctl (s->fd, VIDIOC_QBUF, &buf))
				ms_warning("VIDIOC_QBUF %i failed: %s",k,  strerror(errno));
			else {
				ms_debug("v4l2: queue buf %i",k);
				/*increment ref count of queued buffer*/
				s->frames[k]->b_datap->db_ref++;
				s->queued++;
			}

			if (keyframe==TRUE) {
			  if (s->pix_fmt==MS_H264) {
			    buf.flags = 0;
			    keyframe=FALSE;
			  }
			}
		}
	}

	if (s->queued){
		struct pollfd fds;
		memset(&fds,0,sizeof(fds));
		fds.events=POLLIN;
		fds.fd=s->fd;
		/*check with poll if there is something to read */
		if (poll(&fds,1,0)==1 && fds.revents==POLLIN){
			if (v4l2_ioctl(s->fd, VIDIOC_DQBUF, &buf)<0) {
				switch (errno) {
				case EAGAIN:
				case EIO:
					/* Could ignore EIO, see spec. */
					break;
				default:
					ms_warning("VIDIOC_DQBUF failed: %s",strerror(errno));
				}
			}else{
				s->queued--;
				ms_debug("v4l2: de-queue buf %i",buf.index);
				/*decrement ref count of dequeued buffer */
				ret=s->frames[buf.index];
				ret->b_datap->db_ref--;
				if (buf.index >= s->frame_max){
					ms_error("buf.index>=s->max_frames !");
					return NULL;
				}
				if (buf.bytesused<=4){
					ms_warning("Ignoring empty buffer...");
					return NULL;
				}
				/*normally buf.bytesused should contain the right buffer size; however we have found a buggy
				driver that puts a random value inside */
				if (s->picture_size!=0)
					ret->b_wptr=ret->b_rptr+s->picture_size;
				else ret->b_wptr=ret->b_rptr+buf.bytesused;
			}
		}
	}
	return ret;
}

static void msv4l2_do_munmap(V4l2State *s){
	int i;
	enum v4l2_buf_type type;
	/*stop capture immediately*/
	type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (-1 ==v4l2_ioctl (s->fd, VIDIOC_STREAMOFF, &type)){
		ms_error("VIDIOC_STREAMOFF failed: %s",strerror(errno));
	}

	for(i=0;i<s->frame_max;++i){
		mblk_t *msg=s->frames[i];
		int len=msg->b_datap->db_lim-msg->b_datap->db_base;
		if (v4l2_munmap(msg->b_datap->db_base,len)<0){
			ms_warning("MSV4l2: Fail to unmap: %s",strerror(errno));
		}
		freemsg(s->frames[i]);
		s->frames[i]=NULL;
	}
}



static void msv4l2_init(MSFilter *f){
	V4l2State *s=ms_new0(V4l2State,1);
	s->dev=ms_strdup("/dev/video0");
	s->fd=-1;
	s->vsize=MS_VIDEO_SIZE_CIF;
	s->fps=15;
	s->configured=FALSE;
	s->bitrate=128000;
	s->reset=0;
	f->data=s;
}

static void msv4l2_uninit(MSFilter *f){
	V4l2State *s=(V4l2State*)f->data;
	ms_free(s->dev);
	ms_free(s);
}

static void msv4l2_preprocess(MSFilter *f){
	V4l2State *s=(V4l2State*)f->data;
	if (s->fd==-1 && msv4l2_open(s)!=0) {
	  s->start_time=f->ticker->time;
	  s->frame_count=0;
	  s->reset=0;
		return;
	}
	if (!s->configured && msv4l2_configure(s)!=0){
	  s->start_time=f->ticker->time;
	  s->frame_count=0;
	  s->reset=0;
		return;
	}
	if (msv4l2_do_mmap(s)==0){
		ms_message("V4L2 video capture started.");
	}else{
		msv4l2_close(s);
	}
	s->start_time=f->ticker->time;
	s->frame_count=0;
	s->reset=0;
}

static void msv4l2_process(MSFilter *f){
  V4l2State *d=(V4l2State*)f->data;
  uint32_t elapsed;

  if (d->frame_count==-1){
    d->start_time=f->ticker->time;
    d->frame_count=0;
  }

  
  if ((d->reset==1 && d->fd!=-1)) { // || (d->pix_fmt==MS_H264 && d->frame_count==50 && d->unit==0)) {
    /* reset to integrate new bitrate */
    /* a restart is required for RPI camera d->pix_fmt==MS_H264 && d->unit==0 */
    msv4l2_do_munmap(d);
    msv4l2_close(d);
    msv4l2_preprocess(f);
    ms_message("msv4l2_process: new size & bitrate has been applied (br=%i size=%ix%i)", d->bitrate, d->vsize.width, d->vsize.height);
    if (d->pix_fmt==MS_H264 && d->unit==0)
      d->frame_count=51; /* just avoid a second restart for the raspberry. */
    d->send_idr=FALSE;
  } else if (d->fd!=-1 && d->pix_fmt==MS_H264 && d->unit==0 && (d->frame_count==50 || d->send_idr>0)) {
    int frame_count=d->frame_count;
    d->send_idr=0;
    msv4l2_do_munmap(d);
    msv4l2_close(d);
    msv4l2_preprocess(f);
    ms_message("msv4l2_process: restarted (new IDR requested/wanted)");
    if (frame_count==50)
      d->frame_count=51;
  }

  if (d->fd!=-1){
    
    if (d->pix_fmt==MS_H264) {
#ifdef HAVE_LINUX_UVCVIDEO_H
	mblk_t *m;

	while ((m=v4lv2_grab_image(d, (d->send_idr>0)?TRUE:FALSE))!=NULL) {
	  d->send_idr=0;
	  d->frame_count++;
	  msv4l2_h264_handle(f, m);
	}
#endif
    } else {
      /*see it is necessary to output a frame:*/
      elapsed=f->ticker->time-d->start_time;
      if (((float)elapsed*d->fps/1000.0)>d->frame_count){
	mblk_t *m;
	m=v4lv2_grab_image(d, FALSE);
	if (m){
	  mblk_t *om=dupb(m);
	  mblk_set_marker_info(om,(d->pix_fmt==MS_MJPEG));
	  mblk_set_video_format(om,d->pix_fmt);
	  ms_queue_put(f->outputs[0],om);
	  d->frame_count++;
	}
      }
    }
  }
}

static void msv4l2_postprocess(MSFilter *f){
	V4l2State *s=(V4l2State*)f->data;
	if (s->fd!=-1){
		msv4l2_do_munmap(s);
		msv4l2_close(s);
	}
}

static int msv4l2_set_fps(MSFilter *f, void *arg){
	V4l2State *s=(V4l2State*)f->data;
	s->fps=*(float*)arg;
	s->frame_count=-1;
	return 0;
}

static int msv4l2_set_vsize(MSFilter *f, void *arg){
	V4l2State *s=(V4l2State*)f->data;
	s->vsize=*(MSVideoSize*)arg;
	return 0;
}

static int msv4l2_get_vsize(MSFilter *f, void *arg){
	V4l2State *s=(V4l2State*)f->data;
	*(MSVideoSize*)arg=s->vsize;
	return 0;
}

static int msv4l2_get_pixfmt(MSFilter *f, void *arg){
	V4l2State *s=(V4l2State*)f->data;
	if (s->fd==-1){
		if (msv4l2_open(s)==0){
			msv4l2_configure(s);
			*(MSPixFmt*)arg=s->pix_fmt;
			return 0;
		}else return -1;
	}
	*(MSPixFmt*)arg=s->pix_fmt;
	return 0;
}

static int msv4l2_set_devfile(MSFilter *f, void *arg){
	V4l2State *s=(V4l2State*)f->data;
	if (s->dev) ms_free(s->dev);
	s->dev=ms_strdup((char*)arg);
	return 0;
}

static int msv4l2_set_runtime_br(MSFilter *f, void *arg){
	V4l2State *d=(V4l2State*)f->data;
	int old_br=d->bitrate;
	d->bitrate=*(int*)arg;
	
	if (old_br!=d->bitrate) {
	  ms_message("msv4l2_set_runtime_br: %i", d->bitrate);
	  set_bitrate(d, d->bitrate);
	}
	return 0;
}

static int msv4l2_set_br(MSFilter *f, void *arg){
	V4l2State *d=(V4l2State*)f->data;
	int old_br=d->bitrate;
	int old_w=d->vsize.width;
	int old_h=d->vsize.height;
	d->bitrate=*(int*)arg;
	ms_message("msv4l2_set_br: %i", d->bitrate);

	if (d->bitrate>=1024000){
		d->vsize.width = MS_VIDEO_SIZE_720P_W;
		d->vsize.height = MS_VIDEO_SIZE_720P_H;
		d->fps=30;
	}else if (d->bitrate>=512000){
		d->vsize.width=MS_VIDEO_SIZE_CIF_W;
		d->vsize.height=MS_VIDEO_SIZE_CIF_H;
		d->fps=30;
	}else if (d->bitrate>=256000){
		d->vsize.width=MS_VIDEO_SIZE_CIF_W;
		d->vsize.height=MS_VIDEO_SIZE_CIF_H;
		d->fps=25;
	}else if (d->bitrate>=128000){
		d->vsize.width=MS_VIDEO_SIZE_CIF_W;
		d->vsize.height=MS_VIDEO_SIZE_CIF_H;
		d->fps=20;
	}else if (d->bitrate>=64000){
		d->vsize.width=MS_VIDEO_SIZE_QCIF_W;
		d->vsize.height=MS_VIDEO_SIZE_QCIF_H;
		d->fps=20;
	}else if (d->bitrate==20000){
		d->vsize.width=MS_VIDEO_SIZE_CIF_W;
		d->vsize.height=MS_VIDEO_SIZE_CIF_H;
		d->fps=2;
	}else{
		d->vsize.width=MS_VIDEO_SIZE_QCIF_W;
		d->vsize.height=MS_VIDEO_SIZE_QCIF_H;
		d->fps=20;
	}
	if (d->bitrate!=old_br || d->vsize.width!=old_w || d->vsize.height!=old_h) { d->reset=1; }
	d->max_bitrate=d->bitrate;
	return 0;
}

static int msv4l2_req_vfu(MSFilter *f, void *arg){
  V4l2State *d=(V4l2State*)f->data;
  d->send_idr=1;
  return 0;
}

static MSFilterMethod msv4l2_methods[]={
	{	MS_FILTER_SET_FPS	,	msv4l2_set_fps	},
	{	MS_FILTER_SET_VIDEO_SIZE,	msv4l2_set_vsize	},
	{	MS_FILTER_GET_VIDEO_SIZE,	msv4l2_get_vsize	},
	{	MS_FILTER_GET_PIX_FMT	,	msv4l2_get_pixfmt	},
	{MS_FILTER_SET_BITRATE, msv4l2_set_br },
	{MS_FILTER_SET_RUNTIME_BITRATE, msv4l2_set_runtime_br },
	{MS_FILTER_REQ_VFU, msv4l2_req_vfu},
	{	0			,	NULL		}
};

MSFilterDesc ms_v4l2_desc={
	.id=MS_V4L2_CAPTURE_ID,
	.name="MSV4L2Capture",
	.text=N_("A filter to grab pictures from Video4Linux2-powered cameras"),
	.category=MS_FILTER_OTHER,
	.ninputs=0,
	.noutputs=1,
	.init=msv4l2_init,
	.preprocess=msv4l2_preprocess,
	.process=msv4l2_process,
	.postprocess=msv4l2_postprocess,
	.uninit=msv4l2_uninit,
	.methods=msv4l2_methods
};

MS_FILTER_DESC_EXPORT(ms_v4l2_desc)

static MSFilter *msv4l2_create_reader(MSWebCam *obj){
	MSFilter *f=ms_filter_new(MS_V4L2_CAPTURE_ID);
	msv4l2_set_devfile(f,obj->name);
	return f;
}

static void msv4l2_detect(MSWebCamManager *obj);

static void msv4l2_cam_init(MSWebCam *cam){
}

MSWebCamDesc v4l2_card_desc={
	"V4L2",
	&msv4l2_detect,
	NULL,
	&msv4l2_cam_init,
	&msv4l2_create_reader,
	NULL
};

static void msv4l2_detect(MSWebCamManager *obj){
	struct v4l2_capability cap;
	char devname[32];
	int i;
	for(i=0;i<10;++i){
		int fd;
		snprintf(devname,sizeof(devname),"/dev/video%i",i);
		fd=open(devname,O_RDWR);
		if (fd!=-1){
			if (v4l2_ioctl (fd, VIDIOC_QUERYCAP, &cap)==0) {
				/* is a V4LV2 */
				MSWebCam *cam=ms_web_cam_new(&v4l2_card_desc);
				cam->name=ms_strdup(devname);
				ms_web_cam_manager_add_cam(obj,cam);
			}
			close(fd);
		}
	}
}


#endif
