/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "mediastreamer2/rfc3984.h"
#include "mediastreamer2/msfilter.h"

#define TYPE_FU_A 28    /*fragmented unit 0x1C*/
#define TYPE_FU_B 29    /*fragmented unit 0x1D*/
#define TYPE_STAP_A 24  /*single time aggregation packet  0x18*/
#define TYPE_STAP_B 25  /*single time aggregation packet  0x19*/
#define TYPE_MTAP_16 26  /*single time aggregation packet  0x1A*/
#define TYPE_MTAP_24 27  /*single time aggregation packet  0x1B*/

typedef struct _h264_slice_hdr
{
	uint32_t   first_mb;
	uint16_t slice_type;
} h264_slice_hdr;

static uint32_t get_bit(const uint8_t * const base, uint32_t offset)
{
	return ((*(base + (offset >> 0x3))) >> (0x7 - (offset & 0x7))) & 0x1;
}

static uint32_t get_bits(const uint8_t * const base, uint32_t * const offset, uint8_t bits)
{
	uint32_t value = 0;
	int i;
	for (i = 0; i < bits; i++) {
		value = (value << 1) | (get_bit(base, (*offset)++) ? 1 : 0);
	}
	return value;
}

static uint32_t ue_v(const uint8_t * const base, uint32_t * const offset)
{
	uint32_t zeros = 0;
	uint32_t info;
	int32_t i;
	while (0 == get_bit(base, (*offset)++)) zeros++;

	info = 1 << zeros;

	for (i = zeros - 1; i >= 0; i--) {
		info |= get_bit(base, (*offset)++) << i;
	}

	return (info - 1);
}

static void get_h264_slice_hdr(const uint8_t * const slice_header, h264_slice_hdr *h264_slice_hdr)
{
	uint32_t bit_offset;

	bit_offset = 0;

	h264_slice_hdr->first_mb = ue_v(slice_header, &bit_offset);
	h264_slice_hdr->slice_type = ue_v(slice_header, &bit_offset);
}

static inline void nal_header_init(uint8_t *h, uint8_t nri, uint8_t type){
	*h=((nri&0x3)<<5) | (type & ((1<<5)-1));
}

static inline uint8_t nal_header_get_type(const uint8_t *h){
	return (*h) & ((1<<5)-1);
}

static inline uint8_t nal_header_get_nri(const uint8_t *h){
	return ((*h) >> 5) & 0x3;
}

void rfc3984_init(Rfc3984Context *ctx){
	ms_queue_init(&ctx->q);
	ctx->m=NULL;
	ctx->maxsz=ms_get_payload_max_size();
	ctx->mode=0;
	ctx->last_ts=0x943FEA43;/*some random value*/
	ctx->stap_a_allowed=TRUE;
	ctx->last_cseq=0;
	ctx->nack_lost_detected_cseq=0;
  ctx->nack_lost_bits=0;
  ctx->nack_last_sent_time=0;
  ctx->decoder_ready=FALSE;
	ctx->reserved=0;
}

void rfc3984_set_max_payload_size(Rfc3984Context *ctx, int size){
	ctx->maxsz=size;
}

static void send_packet(MSQueue *rtpq, uint32_t ts, mblk_t *m, bool_t marker){
	mblk_set_timestamp_info(m,ts);
	mblk_set_marker_info(m,marker);
	ms_queue_put(rtpq,m);
}

static void put_nal_size(mblk_t *m, uint16_t sz){
	uint16_t size=htons(sz);
	*(uint16_t*)m->b_wptr=size;
	m->b_wptr+=2;
}

static mblk_t * prepend_stapa(mblk_t *m){
	mblk_t *hm=allocb(3,0);
	nal_header_init(hm->b_wptr,nal_header_get_nri(m->b_rptr),TYPE_STAP_A);
	hm->b_wptr+=1;
	put_nal_size(hm,msgdsize(m));
	hm->b_cont=m;
	return hm;
}

static mblk_t * concat_nalus(mblk_t *m1, mblk_t *m2){
	mblk_t *l=allocb(2,0);
	/*eventually append a stap-A header to m1, if not already done*/
	if (nal_header_get_type(m1->b_rptr)!=TYPE_STAP_A){
		m1=prepend_stapa(m1);
	}
	put_nal_size(l,msgdsize(m2));
	l->b_cont=m2;
	concatb(m1,l);
	return m1;
}

static mblk_t *prepend_fu_indicator_and_header(mblk_t *m, uint8_t indicator,
	bool_t start, bool_t end, uint8_t type){
	mblk_t *h=allocb(2,0);
	h->b_wptr[0]=indicator;
	h->b_wptr[1]=((start&0x1)<<7)|((end&0x1)<<6)|type;
	h->b_wptr+=2;
	h->b_cont=m;
	if (start) m->b_rptr++;/*skip original nalu header */
	return h;
}

static void frag_nalu_and_send(MSQueue *rtpq, uint32_t ts, mblk_t *nalu, bool_t marker, int maxsize){
	mblk_t *m;
	int payload_max_size=maxsize-2;/*minus FUA header*/
	uint8_t fu_indicator;
	uint8_t type=nal_header_get_type(nalu->b_rptr);
	uint8_t nri=nal_header_get_nri(nalu->b_rptr);
	bool_t start=TRUE;

	nal_header_init(&fu_indicator,nri,TYPE_FU_A);
	while(nalu->b_wptr-nalu->b_rptr>payload_max_size){
		m=dupb(nalu);
		nalu->b_rptr+=payload_max_size;
		m->b_wptr=nalu->b_rptr;
		m=prepend_fu_indicator_and_header(m,fu_indicator,start,FALSE,type);
		send_packet(rtpq,ts,m,FALSE);
		start=FALSE;
	}
	/*send last packet */
	m=prepend_fu_indicator_and_header(nalu,fu_indicator,FALSE,TRUE,type);
	send_packet(rtpq,ts,m,marker);
}

static void rfc3984_pack_mode_0(Rfc3984Context *ctx, MSQueue *naluq, MSQueue *rtpq, uint32_t ts){
	mblk_t *m;
	bool_t end;
	int size;
	while((m=ms_queue_get(naluq))!=NULL){
		end=ms_queue_empty(naluq);
		size=(int)(m->b_wptr-m->b_rptr);
		if (size>ctx->maxsz){
			ms_warning("This H264 packet does not fit into mtu: size=%i",size);
		}
		send_packet(rtpq,ts,m,end);
	}
}

/*process NALUs and pack them into rtp payloads */
static void rfc3984_pack_mode_1(Rfc3984Context *ctx, MSQueue *naluq, MSQueue *rtpq, uint32_t ts){
	mblk_t *m,*prevm=NULL;
	int prevsz=0,sz;
	bool_t end;
	while((m=ms_queue_get(naluq))!=NULL){
		end=ms_queue_empty(naluq);
		sz=(int)(m->b_wptr-m->b_rptr);
		if (ctx->stap_a_allowed){
			if (prevm!=NULL){
				if ((prevsz+sz)<(ctx->maxsz-2)){
					prevm=concat_nalus(prevm,m);
					m=NULL;
					prevsz+=sz+2;/*+2 for the stapa size field*/
					continue;
				}else{
					/*send prevm packet: either single nal or STAP-A*/
					if (prevm->b_cont!=NULL){
						ms_debug("Sending STAP-A");
					}else
						ms_debug("Sending previous msg as single NAL");
					send_packet(rtpq,ts,prevm,FALSE);
					prevm=NULL;
					prevsz=0;
				}
			}
			if (sz<(ctx->maxsz/2)){
				/*try to aggregate it with next packet*/
				prevm=m;
				prevsz=sz+3; /*STAP-A header + size*/
				m=NULL;
			}else{
				
				/*send as single nal or FU-A*/
				if (sz>ctx->maxsz){
					ms_debug("Sending FU-A packets");
					frag_nalu_and_send(rtpq,ts,m,end, ctx->maxsz);
				}else{
					ms_debug("Sending Single NAL");
					send_packet(rtpq,ts,m,end);
				}
			}
		}else{
			if (sz>ctx->maxsz){
				ms_debug("Sending FU-A packets");
				frag_nalu_and_send(rtpq,ts,m,end, ctx->maxsz);
			}else{
				ms_debug("Sending Single NAL");
				send_packet(rtpq,ts,m,end);
			}
		}
	}
	if (prevm){
		ms_debug("Sending Single NAL (2)");
		send_packet(rtpq,ts,prevm,TRUE);
	}
}

static mblk_t * aggregate_fua(Rfc3984Context *ctx, mblk_t *im){
	mblk_t *om=NULL;
	uint8_t fu_header;
	uint8_t nri,type;
	bool_t start,end;
	fu_header=im->b_rptr[1];
	type=nal_header_get_type(&fu_header);
	start=fu_header>>7;
	end=(fu_header>>6)&0x1;
	if (start){
		nri=nal_header_get_nri(im->b_rptr);
		if (ctx->m!=NULL){
			ms_error("receiving FU-A start while previous FU-A is not "
				"finished");
			freemsg(ctx->m);
			ctx->m=NULL;
		}
		im->b_rptr++;
		nal_header_init(im->b_rptr,nri,type);
		ctx->m=im;
		if (type == 5) {
			h264_slice_hdr h264_slice_hdr;
			get_h264_slice_hdr(im->b_rptr + 1, &h264_slice_hdr);
			if (h264_slice_hdr.first_mb == 0) {
				/* first_mb==0 means that this is the first slice of an IDR */
				ms_message("rfc3984: IDR with first_mb=%u %i-slice", h264_slice_hdr.first_mb, h264_slice_hdr.slice_type);
				ctx->decoder_ready = TRUE;
			}
		}
	}else{
		if (ctx->m!=NULL){
			im->b_rptr+=2;
			concatb(ctx->m,im);
		}else{
			ms_error("Receiving continuation FU packet but no start.");
			freemsg(im);
		}
	}
	if (end && ctx->m){
		msgpullup(ctx->m,-1);
		om=ctx->m;
		ctx->m=NULL;
	}
	return om;
}

/*process incoming rtp data and output NALUs, whenever possible*/
void rfc3984_unpack(Rfc3984Context *ctx, mblk_t *im, MSQueue *out){
  uint8_t type=nal_header_get_type(im->b_rptr);
  uint8_t *p;
  int marker = mblk_get_marker_info(im);
  uint32_t ts=mblk_get_timestamp_info(im);
  uint16_t cseq=mblk_get_cseq(im);

  if (ctx->last_cseq!=0 && ctx->last_cseq+1!=cseq) {
    mblk_t *old;
    ctx->decoder_ready=FALSE;
    while((old=ms_queue_get(&ctx->q))!=NULL){
      freemsg(old);
    }

    if (ctx->m!=NULL) {
      /* lost packet in the middle of a frame? */
      freemsg(ctx->m);
      ctx->m=NULL;
    }
    if (ctx->last_cseq!=0) {
      if (ctx->nack_lost_detected_cseq>0 && cseq-ctx->nack_lost_detected_cseq>16) {
        /* reset */
        ctx->nack_lost_detected_cseq=0;
        ctx->nack_lost_bits=0;
      }
      if (ctx->nack_lost_detected_cseq==0)
        ctx->nack_lost_detected_cseq = cseq-1;
      else {
        /* complete ctx->nack_lost_bits */
        int pos;
        for (pos=ctx->last_cseq;cseq-pos>1;pos++) {
          ctx->nack_lost_bits |= (1U << (pos-ctx->nack_lost_detected_cseq));
        }
      }
    }
    if (type==TYPE_FU_A) {
      bool_t start = (im->b_rptr[1])>>7;
      if (start==0) {
        /* do not drop if this is the first FU-A packet */
        freemsg(im);
        if (marker) {
          ctx->last_cseq=cseq;
        }
        return;
      }
    }
    if (type==5) {
      /* do not drop if this is the first slice of an IDR picture */
      bool_t start = (im->b_rptr[1])>>7;
      if (start==0) {
        freemsg(im);
        if (marker) {
          ctx->last_cseq=cseq;
        }
        return;
      }
    }

    /* discard everything after up to the packet with a marker bit */
  }
  ctx->last_cseq=cseq;

  if (ctx->last_ts!=ts){
    /*a new frame is arriving, in case the marker bit was not set in previous frame, output it now*/
    /* unless this is a FU-A (workarond some other apps bugs)*/
    ctx->last_ts=ts;
    if (ctx->m==NULL){
      while(!ms_queue_empty(&ctx->q)){
        ms_queue_put(out,ms_queue_get(&ctx->q));
      }
    }
  }

  if (im->b_cont) msgpullup(im,-1);

  if (type==TYPE_STAP_A){
    /*split into nalus*/
    uint16_t sz;
    uint8_t *buf=(uint8_t*)&sz;
    mblk_t *nal;

    for(p=im->b_rptr+1;p<im->b_wptr;){
      buf[0]=p[0];
      buf[1]=p[1];
      sz=ntohs(sz);
      nal=dupb(im);
      p+=2;
      nal->b_rptr=p;
      p+=sz;
      nal->b_wptr=p;
      if (p>im->b_wptr){
        ms_error("Malformed STAP-A packet");
        freemsg(nal);
        break;
      }
	  type = nal_header_get_type(nal->b_rptr);
	  if (type == 5) {
		  h264_slice_hdr h264_slice_hdr;
		  get_h264_slice_hdr(nal->b_rptr + 1, &h264_slice_hdr);
		  if (h264_slice_hdr.first_mb == 0) {
			  /* first_mb==0 means that this is the first slice of an IDR */
			  ms_message("rfc3984: IDR in STAP_A with first_mb=%u %i-slice", h264_slice_hdr.first_mb, h264_slice_hdr.slice_type);
			  ctx->decoder_ready = TRUE;
		  }
	  }
      ms_queue_put(&ctx->q,nal);
    }
    freemsg(im);
  }else if (type==TYPE_FU_A){
    mblk_t *o=aggregate_fua(ctx,im);
    if (o) ms_queue_put(&ctx->q,o);
  }else if (type==TYPE_STAP_B){
    ms_warning("Receiving STAP-B");
    freemsg(im);
  }else if (type==TYPE_FU_B){
    ms_warning("Receiving FU-B");
    freemsg(im);
  }else if (type==TYPE_MTAP_16){
    ms_warning("Receiving MTAP_16");
    freemsg(im);
  }else if (type==TYPE_MTAP_24){
    ms_warning("Receiving MTAP_24");
    freemsg(im);
  }else if (type==30 || type==31 || type==0){
    ms_warning("Receiving undefined nal unit type");
    freemsg(im);
  }else{
    if (ctx->m){
      /*discontinued FU-A, purge it*/
      freemsg(ctx->m);
      ctx->m=NULL;
    }
    if (type==7 || type==8 || type==6){
      if (!marker) {
        /* ms_warning("fake marker = true for sps/pps packet"); */
        marker=1;
      }
    }
	if (type == 5) {
		h264_slice_hdr h264_slice_hdr;
		get_h264_slice_hdr(im->b_rptr + 1, &h264_slice_hdr);
		if (h264_slice_hdr.first_mb == 0) {
			/* first_mb==0 means that this is the first slice of an IDR */
			ms_message("rfc3984: IDR with first_mb=%u %i-slice", h264_slice_hdr.first_mb, h264_slice_hdr.slice_type);
			ctx->decoder_ready = TRUE;
		}
	}
    /*single nal unit*/
    ms_queue_put(&ctx->q,im);
  }

  if (marker){
    ctx->last_ts=ts;
    ms_debug("Marker bit set");
    /*end of frame, output everything*/
    while(!ms_queue_empty(&ctx->q)){
      ms_queue_put(out,ms_queue_get(&ctx->q));
    }
  }
}


void rfc3984_pack(Rfc3984Context *ctx, MSQueue *naluq, MSQueue *rtpq, uint32_t ts){
	switch(ctx->mode){
		case 0:
			rfc3984_pack_mode_0(ctx,naluq,rtpq,ts);
			break;
		case 1:
			rfc3984_pack_mode_1(ctx,naluq,rtpq,ts);
			break;
		default:
			ms_error("Bad or unsupported mode %i",ctx->mode);
	}
}

void rfc3984_uninit(Rfc3984Context *ctx){
	ms_queue_flush(&ctx->q);
	if (ctx->m) freemsg(ctx->m);
	ctx->m=NULL;
}

void rfc3984_set_mode(Rfc3984Context *ctx, int mode){
	ctx->mode=mode;
}

void rfc3984_enable_stap_a(Rfc3984Context *ctx, bool_t yesno){
	ctx->stap_a_allowed=yesno;
}

void rfc3984_analyse_frames(Rfc3984Context *ctx, MSQueue *naluq) {
  mblk_t *im;
  uint8_t *src;
  uint8_t nalu_type;
  queue_t *q = &naluq->q;
  if (ms_queue_empty(naluq))
    return;
  for (im = qbegin(q); !qend(q, im); im = qnext(q, im)) {
    src = im->b_rptr;
    nalu_type = (*src) & ((1 << 5) - 1);

    /* accept incoming SVC / discard other NAL */
    if (nalu_type == 14 || nalu_type == 15 || nalu_type == 20) {
      return;
    }

    if (nalu_type == 5) {
    }
    else if (nalu_type == 1) {
      h264_slice_hdr h264_slice_hdr;
      get_h264_slice_hdr(src + 1, &h264_slice_hdr);
      if (h264_slice_hdr.slice_type == 7) {
        /* If I-slice (= 7) this is equivalent to a key-frame... */
        ms_message("rfc3984: non-IDR with first_mb=%u %i-slice", h264_slice_hdr.first_mb, h264_slice_hdr.slice_type);
        ctx->decoder_ready = TRUE;
      }
    }
    else if (nalu_type == 7)
      ms_message("rfc3984: SPS received");
    else if (nalu_type == 8)
      ms_message("rfc3984: PPS received");
    else {
      ms_message("rfc3984: nalu type=%d", nalu_type);
    }

    while (src<(im->b_wptr - 3)) {
      if (src[0] == 0 && src[1] == 0 && src[2]<3) {
        src += 2;
        /* there might be a NAL here? check its type? */
        //if (!ctx->decoder_ready) {
        //  ms_warning("rfc3984: should report if a nal is there.");
        //}
      }
      src++;
    }
    src++;
    src++;
    src++;
  }
  return;
}
