/*
mediastreamer2 library - modular sound and video processing and streaming
Copyright (C) 2006  Simon MORLAT (simon.morlat@linphone.org)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "test_getopt.h"

#include "mediastreamer2/msfilter.h"
#include "mediastreamer2/mssndcard.h"
#include "mediastreamer2/msticker.h"

#ifndef _WIN32_WCE
#include <signal.h>
#elif defined(_MSC_VER)
#define main _tmain
#endif

static void usage (void);

static void
usage (void)
{
  printf ("echo [options] -r input_rate - input_file [-i ...]\n"
	  "\t--help                  display this help\n"
	  "\t--verbose               set debug level\n"
	  "\t--input-rate            samplerate for input card\n"
	  "\t--output-trate          samplerate for output card\n"
	  "\t--input-alsaperiods     alsa periods for input card\n"
	  "\t--output-alsaperiods    alsa periods for output card\n"
	  "\t--input-alsaperiodsize  alsa periods size for input card\n"
	  "\t--output-alsaperiodsize alsa periods size for output card\n"
	  "\t--input-cardid          select input card name\n"
	  "\t--output-cardid         select output card name\n"
	  );
}

static int run=1;

static void stop(int signum){
	run=0;
}

int main(int argc, const char *const argv[]){
  test_getopt_t *opt;
  int rv;
  const char *optarg;
  int c;
    
	MSFilter *f1,*f2;
	MSSndCard *card_capture;
	MSSndCard *card_playback;
	MSTicker *ticker;
	int input_rate = 16000;
	int output_rate = 16000;
	int input_alsaperiods = 8;
	int output_alsaperiods = 8;
	int input_alsaperiodsize = 256;
	int output_alsaperiodsize = 256;
	char input_cardname[256];
	char output_cardname[256];
	int verbose = ORTP_MESSAGE|ORTP_WARNING|ORTP_ERROR|ORTP_FATAL;
	
	struct test_getopt_option_t long_options[] =
	  {
	    {"help", 'h', 1, "display this help"},
	    {"verbose", 'v', 1, "set debug level"},
	    {"input-rate", 256, 1, "samplerate for input card"},
	    {"output-trate", 257, 1, "samplerate for output card"},
	    {"input-alsaperiods", 258, 1, "alsa periods for input card"},
	    {"output-alsaperiods", 259, 1, "alsa periods for output card"},
	    {"input-alsaperiodsize", 260, 1, "alsa periods size for input card"},
	    {"output-alsaperiodsize", 261, 1, "alsa periods size for output card"},
	    {"input-cardid", 262, 1, "select input card name"},
	    {"output-cardid", 263, 1, "select output card name"},
	    
	    {0, 0, 0, 0}
	  };
	
  	memset(input_cardname, 0, sizeof(input_cardname));
  	memset(output_cardname, 0, sizeof(output_cardname));

	test_getopt_init (&opt, argc, argv);
	
#ifndef _WIN32_WCE
	signal(SIGINT,stop);
#endif

	while ((rv =
		test_getopt_long (opt, long_options, &c, &optarg)) == GETOPT_SUCCESS)
	  {
	    switch (c)
	      {
	      case 'v':
		verbose = atoi (optarg);
		break;
	      case 'h':
	      case 'H':
		printf ("Welcome to the echo application\n");
		printf ("LICENSE is GPL\n");
		exit (0);
		break;
	      case 256:
		input_rate = atoi(optarg);
		break;
	      case 257:
		output_rate = atoi(optarg);
		break;
	      case 258:
		input_alsaperiods = atoi(optarg);
		break;
	      case 259:
		output_alsaperiods = atoi(optarg);
		break;
	      case 260:
		input_alsaperiodsize = atoi(optarg);
		break;
	      case 261:
		output_alsaperiodsize = atoi(optarg);
		break;
	      case 262:
		snprintf(input_cardname, sizeof(input_cardname), "%s", optarg);
		break;
	      case 263:
		snprintf(output_cardname, sizeof(output_cardname), "%s", optarg);
		break;
	      default:
		usage ();
		exit (1);
	      }
          }
	if (rv != GETOPT_EOF)
	  {
	    usage ();
	    exit (1);
	  }

	printf ("debug level:       %i\n", verbose);
	printf ("input-rate:        %i\n", input_rate);
	printf ("output_rate        %i\n", output_rate);
	printf ("input_alsaperiods  %i\n", input_alsaperiods);
	printf ("output_alsaperiods  %i\n", output_alsaperiods);
	printf ("input_alsaperiodsize  %i\n", input_alsaperiodsize);
	printf ("output_alsaperiodsize  %i\n", output_alsaperiodsize);
	printf ("input_cardname  %s\n", input_cardname);
	printf ("output_cardname  %s\n", output_cardname);
	
	ortp_init();
	ortp_set_log_level_mask(verbose);
	ms_init();
	
	if (input_cardname[0]!='\0')
	  card_capture = ms_snd_card_manager_get_card(ms_snd_card_manager_get(),input_cardname);
	else
	  card_capture = ms_snd_card_manager_get_default_capture_card(ms_snd_card_manager_get());
	if (output_cardname[0]!='\0')
	  card_playback = ms_snd_card_manager_get_card(ms_snd_card_manager_get(),output_cardname);
	else
	  card_playback = ms_snd_card_manager_get_default_playback_card(ms_snd_card_manager_get());

#ifdef __linux
	if (card_playback==NULL)
	  card_playback = ms_alsa_card_new_custom(output_cardname, output_cardname);
	if (card_capture==NULL)
	  card_capture = ms_alsa_card_new_custom(input_cardname, input_cardname);
#endif
	
	if (card_playback==NULL || card_capture==NULL){
		ms_error("No card.");
		return -1;
	}
	f1=ms_snd_card_create_reader(card_capture);
	f2=ms_snd_card_create_writer(card_playback);

	ms_filter_call_method (f1, MS_FILTER_SET_SAMPLE_RATE,
			       &input_rate);
	ms_filter_call_method (f1, MS_FILTER_SET_ALSA_PERIODS,
			       &input_alsaperiods);
	ms_filter_call_method (f1, MS_FILTER_SET_ALSA_PERIOD_SIZE,
			       &input_alsaperiodsize);
	
	ms_filter_call_method (f2, MS_FILTER_SET_SAMPLE_RATE,
			       &output_rate);
	ms_filter_call_method (f2, MS_FILTER_SET_ALSA_PERIODS,
			       &output_alsaperiods);
	ms_filter_call_method (f2, MS_FILTER_SET_ALSA_PERIOD_SIZE,
			       &output_alsaperiodsize);

	ticker=ms_ticker_new_withname ("amsip-audio");
	ms_filter_link(f1,0,f2,0);
	ms_ticker_attach(ticker,f1);
#ifndef _WIN32_WCE
	while(run)
		ms_sleep(1);
#else
	ms_sleep(5);
#endif
	ms_ticker_detach(ticker,f1);
	ms_ticker_destroy(ticker);
	ms_filter_unlink(f1,0,f2,0);
	ms_filter_destroy(f1);
	ms_filter_destroy(f2);
	return 0;
}
